#
#   Lightnet postprocessing for Anchor based detectors (Darknet)
#   Copyright EAVISE
#

import logging
import torch
from ..util import BaseTransform

__all__ = ['GetDarknetBoxes', 'GetMultiScaleDarknetBoxes', 'GetBoundingBoxes', 'GetMultiScaleBoundingBoxes']
log = logging.getLogger(__name__)


class GetDarknetBoxes(BaseTransform):
    """ Convert output from darknet networks to bounding box tensor.

    Args:
        conf_thresh (Number [0-1]): Confidence threshold to filter detections
        network_stride (Number): Downsampling factor of the network (most lightnet networks have a `stride` attribute)
        anchors (list): 2D list representing anchor boxes (see :class:`lightnet.models.YoloV2`)

    Returns:
        (Tensor [Boxes x 7]]): **[batch_num, x_tl, y_tl, x_br, y_br, confidence, class_id]** for every bounding box
    """
    def __init__(self, conf_thresh, network_stride, anchors):
        super().__init__()
        self.conf_thresh = torch.tensor(conf_thresh)
        self.network_stride = torch.tensor(network_stride)
        self.anchors = torch.tensor(anchors)
        self.num_anchors = torch.tensor(self.anchors.shape[0])
        self.anchors_step = torch.tensor(self.anchors.shape[1])

    def forward(self, network_output):
        device = network_output.device
        batch, channels, h, w = network_output.shape
        num_classes = (channels // self.num_anchors) - 5

        # Compute xc,yc, w,h, box_score on Tensor
        lin_x = torch.linspace(0, w-1, w).repeat(h, 1).view(h*w).to(device)
        lin_y = torch.linspace(0, h-1, h).view(h, 1).repeat(1, w).view(h*w).to(device)
        anchor_w = self.anchors[:, 0].contiguous().view(1, self.num_anchors, 1).to(device)
        anchor_h = self.anchors[:, 1].contiguous().view(1, self.num_anchors, 1).to(device)

        network_output = network_output.view(batch, self.num_anchors, -1, h*w)          # -1 == 5+num_classes (we can drop feature maps if 1 class)
        network_output[:, :, 0, :].sigmoid_().add_(lin_x).mul_(self.network_stride)     # X center
        network_output[:, :, 1, :].sigmoid_().add_(lin_y).mul_(self.network_stride)     # Y center
        network_output[:, :, 2, :].exp_().mul_(anchor_w).mul_(self.network_stride)      # Width
        network_output[:, :, 3, :].exp_().mul_(anchor_h).mul_(self.network_stride)      # Height
        network_output[:, :, 4, :].sigmoid_()                                           # Box score

        # Compute class_score
        if num_classes > 1:
            with torch.no_grad():
                cls_scores = torch.nn.functional.softmax(network_output[:, :, 5:, :], 2)
            cls_max, cls_max_idx = torch.max(cls_scores, 2)
            cls_max_idx = cls_max_idx.float()
            cls_max.mul_(network_output[:, :, 4, :])
        else:
            cls_max = network_output[:, :, 4, :]
            cls_max_idx = torch.zeros_like(cls_max)

        score_thresh = cls_max > self.conf_thresh
        if score_thresh.sum() == 0:
            return torch.empty(0, 7, device=device)
        else:
            # Mask select boxes > conf_thresh
            coords = network_output.transpose(2, 3)[..., 0:4]
            coords = coords[score_thresh[..., None].expand_as(coords)].view(-1, 4)
            coords = torch.cat([coords[:, 0:2]-coords[:, 2:4]/2, coords[:, 0:2]+coords[:, 2:4]/2], 1)
            scores = cls_max[score_thresh]
            idx = cls_max_idx[score_thresh]

            # Get batch numbers of the detections
            batch_num = score_thresh.view(batch, -1)
            nums = torch.arange(1, batch+1, dtype=torch.uint8, device=batch_num.device)
            batch_num = (batch_num * nums[:, None])[batch_num] - 1

            return torch.cat([batch_num[:, None].float(), coords, scores[:, None], idx[:, None]], dim=1)


def GetBoundingBoxes(*args, **kwargs):
    log.deprecated('GetBoundingBoxes is deprecated, please use the more aptly named "GetDarknetBoxes"')
    log.error('GetDarknetBoxes has different arguments than the previous getBoundingBoxes. Make sure to check the documentation!')
    return GetDarknetBoxes(*args, **kwargs)


class GetMultiScaleDarknetBoxes(GetDarknetBoxes):
    """ Convert the output from multiple yolo output layers (at different scales) to bounding box tensors.

    Args:
        conf_thresh (Number [0-1]): Confidence threshold to filter detections
        network_strides (list): Downsampling factors of the network (most lightnet networks have a `stride` attribute)
        anchors (list): 3D list representing anchor boxes (see :class:`lightnet.models.YoloV3`)

    Returns:
        (Tensor [Boxes x 7]]): **[batch_num, x_tl, y_tl, x_br, y_br, confidence, class_id]** for every bounding box

    Note:
        The `anchors` and `network_strides` should be a list of different values for the different scales.
        When used, this post-processing class calls :class:`~lightnet.data.transform.GetDarknetBoxes` for different scales
        and thus uses different stride and anchors values.

    Warning:
        This post-processing function is not entirely equivalent to the Darknet implementation! |br|
        We just execute the regular :class:`~lightnet.data.transform.GetBoundingBoxes` at multiple scales (different strides and anchors),
        and as such did not implement overlapping class labels.
    """
    def __init__(self, conf_thresh, network_strides, anchors):
        super().__init__(conf_thresh, network_strides[0], anchors[0])
        self.root_strides = network_strides
        self.root_anchors = torch.tensor(anchors, requires_grad=False)

    def forward(self, network_output):
        boxes = []
        for i, output in enumerate(network_output):
            self.network_stride = self.root_strides[i]
            self.anchors = self.root_anchors[i]
            self.num_anchors = self.anchors.shape[0]
            boxes.append(super().forward(output))
        return torch.cat(boxes)


def GetMultiScaleBoundingBoxes(*args, **kwargs):
    log.deprecated('GetMultiScaleBoundingBoxes is deprecated, please use the more aptly named "GetMultiScaleDarknetBoxes"')
    log.error('GetMultiScaleDarknetBoxes has different arguments than the previous getMultiScaleBoundingBoxes. Make sure to check the documentation!')
    return GetMultiScaleDarknetBoxes(*args, **kwargs)



#
#   Darknet YOLOv3 model
#   Copyright EAVISE
#

import functools
from collections import OrderedDict, Iterable
import torch
import torch.nn as nn
import lightnet.network as lnn

__all__ = ['YoloV3']


class YoloV3(lnn.module.Darknet):
    """ Yolo v3 implementation :cite:`yolo_v3`.

    Args:
        num_classes (Number, optional): Number of classes; Default **80**
        input_channels (Number, optional): Number of input channels; Default **3**
        anchors (list, optional): 3D list with anchor values; Default **Yolo v3 anchors (COCO)**

    Attributes:
        self.stride: Subsampling factors of the network (input_dim / output_dim)
        self.inner_stride: Maximal internal subsampling factor of the network (input dimension should be a multiple of this)
        self.remap_darknet53: Remapping rules for weights from the :class:`~lightnet.models.Darknet53` model.

    Note:
        Unlike YoloV2, the anchors here are defined as multiples of the input dimensions and not as a multiple of the output dimensions!
        The anchor list also has one more dimension than the one from YoloV2, in order to differentiate which anchors belong to which stride.

    Warning:
        The :class:`~lightnet.network.loss.MultiScaleRegionLoss` and :class:`~lightnet.data.transform.GetMultiScaleBoundingBoxes`
        do not implement the overlapping class labels of the original implementation.
        Your weight files from darknet will thus not have the same accuracies as in darknet itself.
    """
    stride = (32, 16, 8)
    inner_stride = 32
    remap_darknet53 = [
        (r'^layers.([a-w]_)',   r'extractor.\1'),   # Residual layers
        (r'^layers.(\d_)',      r'extractor.\1'),   # layers 1, 2, 5
        (r'^layers.([124]\d_)', r'extractor.\1'),   # layers 10, 27, 44
    ]

    def __init__(self, num_classes, input_channels=3, anchors=[[(116, 90), (156, 198), (373, 326)], [(30, 61), (62, 45), (59, 119)], [(10, 13), (16, 30), (33, 23)]]):
        super().__init__()
        if not isinstance(anchors, Iterable) and not isinstance(anchors[0], Iterable) and not isinstance(anchors[0][0], Iterable):
            raise TypeError('Anchors need to be a 3D list of numbers')

        # Parameters
        self.num_classes = num_classes
        self.anchors = []   # YoloV3 defines anchors as a multiple of the input dimensions of the network as opposed to the output dimensions
        for i, s in enumerate(self.stride):
            self.anchors.append([(a[0] / s, a[1] / s) for a in anchors[i]])

        # Network
        relu = functools.partial(nn.LeakyReLU, 0.1, inplace=True)
        momentum = 0.01
        self.extractor = lnn.layer.SequentialSelect(
            ['k_residual', 's_residual'], True,
            OrderedDict([
                ('1_convbatch',         lnn.layer.Conv2dBatchReLU(input_channels, 32, 3, 1, 1, relu=relu, momentum=momentum)),
                ('2_convbatch',         lnn.layer.Conv2dBatchReLU(32, 64, 3, 2, 1, relu=relu, momentum=momentum)),
                ('a_residual',          lnn.layer.Residual(OrderedDict([
                    ('3_convbatch',     lnn.layer.Conv2dBatchReLU(64, 32, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('4_convbatch',     lnn.layer.Conv2dBatchReLU(32, 64, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('5_convbatch',         lnn.layer.Conv2dBatchReLU(64, 128, 3, 2, 1, relu=relu, momentum=momentum)),
                ('b_residual',          lnn.layer.Residual(OrderedDict([
                    ('6_convbatch',     lnn.layer.Conv2dBatchReLU(128, 64, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('7_convbatch',     lnn.layer.Conv2dBatchReLU(64, 128, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('c_residual',          lnn.layer.Residual(OrderedDict([
                    ('8_convbatch',     lnn.layer.Conv2dBatchReLU(128, 64, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('9_convbatch',     lnn.layer.Conv2dBatchReLU(64, 128, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('10_convbatch',        lnn.layer.Conv2dBatchReLU(128, 256, 3, 2, 1, relu=relu, momentum=momentum)),
                ('d_residual',          lnn.layer.Residual(OrderedDict([
                    ('11_convbatch',    lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('12_convbatch',    lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('e_residual',          lnn.layer.Residual(OrderedDict([
                    ('13_convbatch',    lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('14_convbatch',    lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('f_residual',          lnn.layer.Residual(OrderedDict([
                    ('15_convbatch',    lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('16_convbatch',    lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('g_residual',          lnn.layer.Residual(OrderedDict([
                    ('17_convbatch',    lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('18_convbatch',    lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('h_residual',          lnn.layer.Residual(OrderedDict([
                    ('19_convbatch',    lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('20_convbatch',    lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('i_residual',          lnn.layer.Residual(OrderedDict([
                    ('21_convbatch',    lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('22_convbatch',    lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('j_residual',          lnn.layer.Residual(OrderedDict([
                    ('23_convbatch',    lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('24_convbatch',    lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('k_residual',          lnn.layer.Residual(OrderedDict([
                    ('25_convbatch',    lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('26_convbatch',    lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('27_convbatch',        lnn.layer.Conv2dBatchReLU(256, 512, 3, 2, 1, relu=relu, momentum=momentum)),
                ('l_residual',          lnn.layer.Residual(OrderedDict([
                    ('28_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('29_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('m_residual',          lnn.layer.Residual(OrderedDict([
                    ('30_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('31_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('n_residual',          lnn.layer.Residual(OrderedDict([
                    ('32_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('33_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('o_residual',          lnn.layer.Residual(OrderedDict([
                    ('34_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('35_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('p_residual',          lnn.layer.Residual(OrderedDict([
                    ('36_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('37_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('q_residual',          lnn.layer.Residual(OrderedDict([
                    ('38_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('39_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('r_residual',          lnn.layer.Residual(OrderedDict([
                    ('40_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('41_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('s_residual',          lnn.layer.Residual(OrderedDict([
                    ('42_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('43_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('44_convbatch',        lnn.layer.Conv2dBatchReLU(512, 1024, 3, 2, 1, relu=relu, momentum=momentum)),
                ('t_residual',          lnn.layer.Residual(OrderedDict([
                    ('45_convbatch',    lnn.layer.Conv2dBatchReLU(1024, 512, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('46_convbatch',    lnn.layer.Conv2dBatchReLU(512, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('u_residual',          lnn.layer.Residual(OrderedDict([
                    ('47_convbatch',    lnn.layer.Conv2dBatchReLU(1024, 512, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('48_convbatch',    lnn.layer.Conv2dBatchReLU(512, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('v_residual',          lnn.layer.Residual(OrderedDict([
                    ('49_convbatch',    lnn.layer.Conv2dBatchReLU(1024, 512, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('50_convbatch',    lnn.layer.Conv2dBatchReLU(512, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('w_residual',          lnn.layer.Residual(OrderedDict([
                    ('51_convbatch',    lnn.layer.Conv2dBatchReLU(1024, 512, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('52_convbatch',    lnn.layer.Conv2dBatchReLU(512, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
            ]),
        )

        self.detector = nn.ModuleList([
            # Sequence 0 : input = extractor
            lnn.layer.SequentialSelect(
                ['57_convbatch'], True,
                OrderedDict([
                    ('53_convbatch',    lnn.layer.Conv2dBatchReLU(1024, 512, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('54_convbatch',    lnn.layer.Conv2dBatchReLU(512, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                    ('55_convbatch',    lnn.layer.Conv2dBatchReLU(1024, 512, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('56_convbatch',    lnn.layer.Conv2dBatchReLU(512, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                    ('57_convbatch',    lnn.layer.Conv2dBatchReLU(1024, 512, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('58_convbatch',    lnn.layer.Conv2dBatchReLU(512, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                    ('59_conv',         nn.Conv2d(1024, len(self.anchors[0])*(5+self.num_classes), 1, 1, 0)),
                ])
            ),

            # Sequence 1 : input = 57_convbatch
            nn.Sequential(
                OrderedDict([
                    ('60_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('61_upsample',     nn.Upsample(scale_factor=2, mode='nearest')),
                ])
            ),

            # Sequence 2 : input = 61_upsample and s_residual
            lnn.layer.SequentialSelect(
                ['66_convbatch'], True,
                OrderedDict([
                    ('62_convbatch',    lnn.layer.Conv2dBatchReLU(256+512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('63_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                    ('64_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('65_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                    ('66_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('67_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                    ('68_conv',         nn.Conv2d(512, len(self.anchors[1])*(5+self.num_classes), 1, 1, 0)),
                ])
            ),

            # Sequence 3 : input = 66_convbatch
            nn.Sequential(
                OrderedDict([
                    ('69_convbatch',    lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('70_upsample',     nn.Upsample(scale_factor=2, mode='nearest')),
                ])
            ),

            # Sequence 4 : input = 70_upsample and k_residual
            nn.Sequential(
                OrderedDict([
                    ('71_convbatch',    lnn.layer.Conv2dBatchReLU(128+256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('72_convbatch',    lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                    ('73_convbatch',    lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('74_convbatch',    lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                    ('75_convbatch',    lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('76_convbatch',    lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                    ('77_conv',         nn.Conv2d(256, len(self.anchors[2])*(5+self.num_classes), 1, 1, 0)),
                ])
            ),
        ])

    def forward(self, x):
        # Feature extractor
        x, inter_features = self.extractor(x)

        # Detector 0
        out_0, x = self.detector[0](x)

        # Detector 1
        x = self.detector[1](x)
        out_1, x = self.detector[2](torch.cat((x, inter_features['s_residual']), 1))

        # Detector 2
        x = self.detector[3](x)
        out_2 = self.detector[4](torch.cat((x, inter_features['k_residual']), 1))

        return (out_0, out_1, out_2)



#
#   Lightnet related data processing
#   Utilitary classes and functions for the data subpackage
#   Copyright EAVISE
#

import logging
from abc import ABC, abstractmethod
import numpy as np
import torch
from .._imports import pd, Image

__all__ = ['Compose']
log = logging.getLogger(__name__)


class BaseTransform(ABC, torch.nn.Module):
    """ Base transform class for the pre- and post-processing functions.
    This class allows to create an object with some case specific settings, and then call it with the data to perform the transformation.
    It also allows to call the static method ``apply()`` with the data and settings. This is usefull if you want to transform a single data object.

    There are 2 ways to use this base class.
        - You provide implementations for ``_tf_pil()``, ``_tf_cv()`` and ``_tf_torch()`` to transform the different image types.
        - You override the ``forward()`` method and provide your own data transformation pipeline.
    """
    def forward(self, data):
        if data is None:
            return None
        elif isinstance(data, torch.Tensor):
            return self._tf_torch(data)
        elif Image is not None and isinstance(data, Image.Image):
            return self._tf_pil(data)
        elif isinstance(data, np.ndarray):
            return self._tf_cv(data)
        else:
            log.error(f'{self.__class__.__name__} only works with <brambox dataframes>, <PIL images>, <OpenCV images> or <torch Tensors> [{type(data)}]')
            return data

    @classmethod
    def apply(cls, data, **kwargs):
        """ Classmethod that applies the transformation once.

        Args:
            data: Data to transform (eg. image)
            **kwargs: Same arguments that are passed to the ``__init__`` function
        """
        obj = cls(**kwargs)
        return obj(data)

    def _tf_pil(self, img):
        raise NotImplementedError('This transformation is not implemented for PIL images.')

    def _tf_cv(self, img):
        raise NotImplementedError('This transformation is not implemented for OpenCV NumPy images.')

    def _tf_torch(self, img):
        raise NotImplementedError('This transformation is not implemented for PyTorch Tensor images.')

    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        string = f'{self.__class__.__name__} (\n'

        for name in sorted(self.__dict__.keys()):
            if name.startswith('_'):
                continue
            val = self.__dict__[name]

            valrepr = repr(val)
            if '\n' in valrepr:
                valrepr = val.__class__.__name__

            string += f'  {name} = {valrepr},\n'

        return string + ')'


class BaseMultiTransform(ABC, torch.nn.Module):
    """ Base multiple transform class that is mainly used in pre-processing functions.
    This class exists for transforms that affect both images and annotations.
    It provides a classmethod ``apply``, that will perform the transormation on one (data, target) pair.

    In order to use this base class, you need to provide implementations for ``_tf_pil()``, ``_tf_cv()`` and ``_tf_torch()`` to transform the different image types
    and ``_tf_anno()`` to transform the annotations.
    """
    def forward(self, data):
        if data is None:
            return None
        elif isinstance(data, torch.Tensor):
            return self._tf_torch(data)
        elif pd is not None and isinstance(data, pd.DataFrame):
            return self._tf_anno(data)
        elif Image is not None and isinstance(data, Image.Image):
            return self._tf_pil(data)
        elif isinstance(data, np.ndarray):
            return self._tf_cv(data)
        else:
            log.error(f'{self.__class__.__name__} only works with <brambox dataframes>, <PIL images>, <OpenCV images> or <torch Tensors> [{type(data)}]')
            return data

    @classmethod
    def apply(cls, data, target=None, **kwargs):
        """ Classmethod that applies the transformation once.

        Args:
            data: Data to transform (eg. image)
            target (optional): ground truth for that data; Default **None**
            **kwargs: Same arguments that are passed to the ``__init__`` function
        """
        obj = cls(**kwargs)
        res_data = obj(data)

        if target is None:
            return res_data

        res_target = obj(target)
        return res_data, res_target

    @abstractmethod
    def _tf_pil(self, img):
        return img

    @abstractmethod
    def _tf_cv(self, img):
        return img

    @abstractmethod
    def _tf_torch(self, img):
        return img

    @abstractmethod
    def _tf_anno(self, anno):
        return anno

    def __str__(self):
        return f'{self.__class__.__name__}'

    def __repr__(self):
        string = f'{self.__class__.__name__} [MULTI-TF] (\n'

        for name in sorted(self.__dict__.keys()):
            if name.startswith('_'):
                continue
            val = self.__dict__[name]

            valrepr = repr(val)
            if '\n' in valrepr:
                valrepr = val.__class__.__name__

            string += f'  {name} = {valrepr},\n'

        return string + ')'


class Compose(list):
    """ This is lightnet's own version of :class:`torchvision.transforms.Compose`, which has some extra bells and whistles.

    One of its main features is that you can create a single pipeline for both your images and annotations.
    When you want to run your pipeline, you simply call this compose list with a tuple,
    containing your image and your annotations (annotation has to be :class:`pandas.DataFrame`).
    This class will then run through the transformations and apply each of them.
    If a transformation is of a type :class:`~lightnet.data.transform.util.BaseMultiTransform`,
    all elements from the data tuple will be run through this transformation sequentially,
    otherwise only the first item will be transformed. |br|
    If you pass anything other than a tuple, it will just be transformed by each transformation sequentially.

    Check out the `tutorial <../notes/02-A-basics.html#Pre-processing-pipeline>`_ for more information.

    Args:
        transformations (list of callables): A list of all your transformations in the right order.

    Attributes:
        self.multi_tf (tuple): Which classes to consider to be multi-transforms that act on both images and annotations; Default **(BaseMultiTransform,)**

    Example:
        Adding and removing transformations on the fly, using list methods:

        >>> tf = ln.data.transform.Compose([lambda n: n+1])
        >>> tf(10)  # 10+1
        11
        >>> # We can append using the append method
        >>> tf.append(lambda n: n*2)
        >>> tf(10)  # (10+1)*2
        22
        >>> # Because our composition class is a list, we can also append by adding a list
        >>> tf += [lambda n: n**2]
        >>> tf(10)  # ((10+1)*2)**2
        484
        >>> # Inserting at a random place in the list
        >>> tf.insert(0, lambda n: n//2)
        >>> tf(10)  # (((10//2)+1)*2)**2
        144
        >>> # Removing an operator
        >>> del tf[2]
        >>> tf(10)  # ((10//2)+1)**2
        36

        Combining pipelines:

        >>> pipeline1 = ln.data.transform.Compose([lambda n: n+1])
        >>> pipeline1(10)   # 10+1
        11
        >>> pipeline2 = ln.data.transform.Compose([lambda n: n-1])
        >>> pipeline2(10)   # 10-1
        9
        >>> pipeline = pipeline1 + pipeline2
        >>> pipeline(10)    # (10+1)-1
        10

        Named indexes:

        >>> pipeline = ln.data.transform.Compose([
        ...     ln.data.transform.RandomHSV(hue=1, saturation=2, value=2),
        ...     ln.data.transform.Letterbox(dimension=(416, 416)),
        ... ])
        >>> print(pipeline)
        Compose [RandomHSV, Letterbox]
        >>> # Access transforms by index
        >>> print(pipeline[1])
        Letterbox
        >>> # Access transforms by name (casing does not matter)
        >>> print(pipeline['randomhsv'])
        RandomHSV
        >>> # Check if a certain transformation is in this pipeline
        >>> 'letterbox' in pipeline
        True
        >>> 'RandomCrop' in pipeline
        False
    """
    multi_tf = (BaseMultiTransform,)

    def __call__(self, data):
        """ Run your data through the transformation pipeline.

        Args:
            data: The data to modify. If it is a tuple, only the first item will be transformed, unless the transform is an instance of self.multi_tf.
        """
        if isinstance(data, tuple) and any(isinstance(d, pd.DataFrame) for d in data):
            for tf in self:
                if isinstance(tf, self.multi_tf):
                    data = tuple(tf(d) for d in data)
                else:
                    data = tuple(d if isinstance(d, pd.DataFrame) else tf(d) for d in data)
        else:
            for tf in self:
                data = tf(data)

        return data

    def __getitem__(self, index):
        """ Get a specific item from the transformation list.

        If the index is a string, we compare this string with the class or function names of the transformations in the list (all lowercase).
        For classes, we use `tf.__class__.__name__.lower()`, otherwise we use `tf.__name__.lower()`.
        If there are multiple transforms from the same class, we return the first match.

        If the index is not a string, we simply call the ``__getitem__()`` method from list, which expects an integer.
        """
        if isinstance(index, str):
            index = index.lower()
            keys = tuple(tf.__class__.__name__.lower() if tf.__class__.__name__ != 'function' else tf.__name__.lower() for tf in self)
            if index not in keys:
                raise KeyError(f'[{index}] not found in transforms')

            return super().__getitem__(keys.index(index))
        else:
            return super().__getitem__(index)

    def __contains__(self, key):
        """ Check if a class or function is in this compose list.
        If the key is a string, we compare it with the class or function names of the transformations in this list (all lowercase).
        For classes, we use `tf.__class__.__name__.lower()`, otherwise we use `tf.__name__.lower()`.
        """
        if isinstance(key, str):
            keys = tuple(tf.__class__.__name__.lower() if tf.__class__.__name__ != 'function' else tf.__name__.lower() for tf in self)
            return key.lower() in keys
        else:
            return super().__getitem__(index)

    def __str__(self):
        string = f'{self.__class__.__name__} ['
        for tf in self:
            string += f'{str(tf)}, '
        return string[:-2] + ']'

    def __repr__(self):
        format_string = self.__class__.__name__ + ' ['
        for tf in self:
            tfrepr = repr(tf)
            if '\n' in tfrepr:
                tfrepr = tfrepr.replace('\n', '\n  ')
            format_string += f'\n  {tfrepr}'
        format_string += '\n]'
        return format_string

    def __add__(self, other):
        return Compose(super().__add__(other))

    def __radd__(self, other):
        return Compose(other.__add__(self))

    def __mul__(self, other):
        return Compose(super().__mul__(other))

    def __rmul__(self, other):
        return Compose(other.__mul__(self))



#
#   Base engine class
#   Copyright EAVISE
#

import logging
import signal
from abc import ABC, abstractmethod

import lightnet as ln

__all__ = ['Engine']
log = logging.getLogger(__name__)


class Engine(ABC):
    """ This class removes the boilerplate code needed for writing your training cycle. |br|

    .. warning::
       There are already a lot of PyTorch libraries that are created to ease the creation of a training pipeline.

       In order to limit the burden on the Lightnet dev-team, we will stop working on our own engine
       and instead be slowly transitioning towards `PyTorch Lightning <lightning_>`_. |br|
       This transition will be slow and this engine will thus remain in the codebase for quite some time,
       but no further development will be made to this.

       Besides, PyTorch Lightnig offers a ton of extra functionality
       and is being maintained by a much bigger group of people,
       which allows it to stay up-to-date with recent Deep Learning trends much faster!

    Here is the code that runs when the engine is called:

    .. literalinclude:: /../lightnet/engine/_engine.py
       :language: python
       :pyobject: Engine.__call__
       :dedent: 4

    Args:
        params (lightnet.engine.HyperParameters): Serializable hyperparameters for the engine to work with
        dataloader (torch.utils.data.DataLoader, optional): Dataloader for the training data; Default **None**
        **kwargs (dict, optional): Keywords arguments that will be set as attributes of the engine

    Attributes:
        self.params: HyperParameter object
        self.dataloader: Dataloader object
        self.sigint: Boolean value indicating whether a SIGINT (CTRL+C) was send; Default **False**
        self.*: All values that were passed with the init function and all values from the :class:`~lightnet.engine.HyperParameters` can be accessed in this class

    Note:
        This class expects a `self.dataloader` object to be present. |br|
        You can either pass a dataloader when initializing this class, or you can define it yourself.
        This allows to define `self.dataloader` as a computed property (@property) of your class, opening up a number of different possibilities,
        like eg. computing different dataloaders depending on which epoch you are.

    Note:
        This engine allows to define hook functions to run at certain points in the training *(epoch_start, epoch_end, batch_start, batch_end)*.
        The functions can be defined as class methods of your engine without any extra arguments or as separate functions that take the engine as a single argument.

        There are different functions to register a hook and they can be used as decorator functions or called straight away in code:

        >>> class TrainingEngine(ln.engine.Engine):
        ...     def start(self):
        ...         pass
        ...
        ...     @ln.engine.Engine.epoch_end
        ...     def backup(self):
        ...         pass    # This method will be executed at the end of every epoch
        ...
        ...     @ln.engine.Engine.batch_start(100)
        ...     def update_hyperparams(self):
        ...         pass    # This method will be executed at the start of every 100th batch
        ...
        >>> # Create TrainingEngine object and run it

        >>> def backup(engine):
        ...     pass    # This function will be executed at the end of every Xth batch defined by a backup_rate variable at runtime
        ...
        >>> @ln.engine.Engine.epoch_start
        ... def select_data_subset(engine):
        ...     pass    # This function will be executed at the start of every epoch
        ...
        >>> class TrainingEngine(ln.engine.Engine):
        ...     def start(self):
        ...         if hasattr(self, 'backup_rate') and self.backup_rate is not None:
        ...             self.batch_start(self.backup_rate)(backup)
        ...
        >>> # Create TrainingEngine object and run it
    """
    __init_done = False
    _required_attr = ['network', 'batch_size', 'dataloader']
    _handled_signals = [signal.SIGINT, signal.SIGTERM]
    _epoch_start = {}
    _epoch_end = {}
    _batch_start = {}
    _batch_end = {}

    def __init__(self, params, dataloader=None, **kwargs):
        self.params = params
        if dataloader is not None:
            self.dataloader = dataloader

        # Sigint handling
        self.sigint = False
        for sig in self._handled_signals:
            signal.signal(sig, self.__sigint_handler)

        # Set attributes
        for key in kwargs:
            if not hasattr(self, key):
                setattr(self, key, kwargs[key])
            else:
                log.warning(f'{key} attribute already exists on engine.')

        self.__init_done = True

    def __call__(self):
        """ Start the training cycle. """
        self.__check_attr()
        self.start()

        log.info('Start training')
        self.network.train()

        idx = 0
        while True:
            # Check if we need to stop training
            if self.quit() or self.sigint:
                log.info('Reached quitting criteria')
                return

            # Epoch Start
            self._run_hooks(self.epoch + 1, self._epoch_start)

            idx %= self.batch_subdivisions
            loader = self.dataloader
            for idx, data in enumerate(loader, idx+1):
                # Batch Start
                if (idx - 1) % self.batch_subdivisions == 0:
                    self._run_hooks(self.batch + 1, self._batch_start)

                # Forward and backward on (mini-)batches
                self.process_batch(data)
                if idx % self.batch_subdivisions != 0:
                    continue

                # Optimizer step
                self.batch += 1     # Should only be called after train, but this is easier to use self.batch in function
                self.train_batch()

                # Batch End
                self._run_hooks(self.batch, self._batch_end)

                # Check if we need to stop training
                if self.quit() or self.sigint:
                    log.info('Reached quitting criteria')
                    return

            # Epoch End
            self.epoch += 1
            self._run_hooks(self.epoch, self._epoch_end)

    def __getattr__(self, name):
        if hasattr(self.params, name):
            return getattr(self.params, name)
        else:
            raise AttributeError(f'{name} attribute does not exist')

    def __setattr__(self, name, value):
        if self.__init_done and name not in dir(self) and hasattr(self.params, name):
            setattr(self.params, name, value)
        else:
            super().__setattr__(name, value)

    def __sigint_handler(self, signal, frame):
        if not self.sigint:
            log.debug('SIGINT/SIGTERM caught. Waiting for gracefull exit')
            self.sigint = True

    def __check_attr(self):
        for attr in self._required_attr:
            if not hasattr(self, attr):
                raise AttributeError(f'Engine requires attribute [{attr}] (as an engine or hyperparameter attribute)')

        if not hasattr(self, 'mini_batch_size'):
            log.warning('No [mini_batch_size] attribute found, setting it to [batch_size]')
            self.mini_batch_size = self.batch_size
        elif self.batch_size % self.mini_batch_size != 0 or self.mini_batch_size > self.batch_size:
            raise ValueError('batch_size should be a multiple of mini_batch_size')

    def log(self, msg):
        """ Log messages about training and testing.
        This function will automatically prepend the messages with **TRAIN** or **TEST**.

        Args:
            msg (str): message to be printed
        """
        if self.network.training:
            log.train(msg)
        else:
            log.test(msg)

    def _run_hooks(self, value, hooks):
        """ Internal method that will execute registered hooks. """
        keys = list(hooks.keys())
        for k in keys:
            if value % k == 0:
                for fn in hooks[k]:
                    if hasattr(fn, '__self__'):
                        fn()
                    else:
                        fn(self)

    @classmethod
    def epoch_start(cls, interval=1):
        """ Register a hook to run at the start of an epoch.

        Args:
            interval (int, optional): Number dictating how often to run the hook; Default **1**

        Note:
            The `self.epoch` attribute contains the number of processed epochs,
            and will thus be one lower than the epoch you are currently starting.
            For example, when starting training with the very first epoch,
            the `self.epoch` attribute will be set to **0** during any `epoch_start` hook. |br|
            However, the `interval` argument will be computed with the correct epoch number (ic. `self.epoch` + 1).
        """
        def decorator(fn):
            if interval in cls._epoch_start:
                cls._epoch_start[interval].append(fn)
            else:
                cls._epoch_start[interval] = [fn]
            return fn

        return decorator

    @classmethod
    def epoch_end(cls, interval=1):
        """ Register a hook to run at the end of an epoch.

        Args:
            interval (int, optional): Number dictating how often to run the hook; Default **1**
        """
        def decorator(fn):
            if interval in cls._epoch_end:
                cls._epoch_end[interval].append(fn)
            else:
                cls._epoch_end[interval] = [fn]
            return fn

        return decorator

    @classmethod
    def batch_start(cls, interval=1):
        """ Register a hook to run at the start of a batch.

        Args:
            interval (int, optional): Number dictating how often to run the hook; Default **1**

        Note:
            The `self.batch` attribute contains the number of processed batches,
            and will thus be one lower than the batch you are currently starting.
            For example, when starting training with the very first batch,
            the `self.batch` attribute will be set to **0** during any `batch_start` hook. |br|
            However, the `interval` argument will be computed with the correct batch number (ic. `self.batch` + 1).
        """
        def decorator(fn):
            if interval in cls._batch_start:
                cls._batch_start[interval].append(fn)
            else:
                cls._batch_start[interval] = [fn]
            return fn

        return decorator

    @classmethod
    def batch_end(cls, interval=1):
        """ Register a hook to run at the end of a batch.

        Args:
            interval (int, optional): Number dictating how often to run the hook; Default **1**
        """
        def decorator(fn):
            if interval in cls._batch_end:
                cls._batch_end[interval].append(fn)
            else:
                cls._batch_end[interval] = [fn]
            return fn

        return decorator

    @property
    def batch_subdivisions(self):
        """ Get number of mini-batches per batch.

        Return:
            int: Computed as self.batch_size // self.mini_batch_size
        """
        return self.batch_size // self.mini_batch_size

    def start(self):
        """ First function that gets called when starting the engine. |br|
        Any required setup code can come in here.
        """
        pass

    @abstractmethod
    def process_batch(self, data):
        """ This function should contain the code to process the forward and backward pass of one (mini-)batch.

        Args:
            data: The data that comes from your dataloader

        Note:
            If you are working with mini-batches, you should pay attention to how you process your loss and backwards function. |br|
            PyTorch accumulates gradients when performing multiple backward() calls before using your optimizer.
            However, usually your loss function performs some kind of average over your batch-size (eg. reduction='mean' in a lot of default pytorch functions).
            When that is the case, you should also average your losses over the mini-batches, by dividing your resulting loss:

            .. code:: bash

                loss = loss_function(output, target) / self.batch_subdivisions
                loss.backward()
        """
        pass

    @abstractmethod
    def train_batch(self):
        """ This function should contain the code to update the weights of the network. |br|
        Statistical computations, performing backups at regular intervals, etc. also happen here.
        """
        pass

    def quit(self):
        """ This function gets called after every training epoch and decides if the training cycle continues.

        Return:
            Boolean: Whether are not to stop the training cycle

        Note:
            This function gets called before checking the ``self.sigint`` attribute.
            This means you can also check this attribute in this function. |br|
            If it evaluates to **True**, you know the program will exit after this function and you can thus
            perform the necessary actions (eg. save final weights).
        """
        return False



#
#   Learning Rate Scheduling
#   Copyright EAVISE
#

import torch
import logging

__all__ = ['SchedulerCompositor']
log = logging.getLogger(__name__)


class SchedulerCompositor:
    """ This class can be used to schedule schedulers to run at different moments on the same parameters of a network. |br|
    This compositor has a notion of count values. These can be batch number, epoch number, etc. and dictate when each scheduler is being used.

    Args:
        *args (list of tuples): This class gets initialized with tuples of ``(count, scheduler)``, which determine when to start using which scheduler

    Example:
        >>> class DummyScheduler:
        ...     " Dummy scheduler that does nothing but print it's id value. "
        ...     def __init__(self, id_value):
        ...         self.id = id_value;
        ...     def step(self):
        ...         print(f'{self.last_epoch} - Dummy Scheduler: {self.id}')
        >>> s = ln.engine.SchedulerCompositor(
        ...     (1, DummyScheduler('start')),
        ...     (3, DummyScheduler('middle')),
        ...     (4, DummyScheduler('end')),
        ... )
        >>> # Note that then "count_value" argument is just for this example
        >>> for i in range(5):
        ...     s.step()
        1 - Dummy Scheduler: start
        2 - Dummy Scheduler: start
        3 - Dummy Scheduler: middle
        4 - Dummy Scheduler: end
        5 - Dummy Scheduler: end
    """
    def __init__(self, *args, last_epoch=0):
        if len(args) == 0:
            raise ValueError('Compositor requires at least one scheduler')

        self.epoch, self.sched = zip(*args)
        if not all(c1 < c2 for c1, c2 in zip(self.epoch, self.epoch[1:])):
            raise ValueError('Count values need to be strictly increasing')

        self.last_epoch = last_epoch

    def __repr__(self):
        format_string = self.__class__.__name__ + ' ['
        clen = max(len(str(c)) for c in self.epoch)
        for i in range(len(self.epoch)):
            if hasattr(self.sched[i], '__name__'):
                name = self.sched[i].__name__
            else:
                name = self.sched[i].__class__.__name__

            format_string += f'\n  {self.epoch[i]:>{clen}}:  {name}'
        format_string += '\n]'
        return format_string

    def get_last_lr(self):
        sched = self._get_scheduler(self.last_epoch)
        if sched is not None:
            return sched.get_last_lr()

    def step(self, **kwargs):
        """ Stepping function that will select a scheduler and run it.

        Args:
            **kwargs (dict, optional): Extra arguments that will be passed on to the step function of the scheduler itself.
        """
        self.last_epoch += 1
        sched = self._get_scheduler(self.last_epoch)

        if sched is not None:
            sched.last_epoch = self.last_epoch
            return sched.step(**kwargs)

    def _get_scheduler(self, epoch):
        for i, e in enumerate(self.epoch):
            if epoch < e:
                i -= 1
                break

        if i < 0:
            log.error(f'No Scheduler defined for count value of {epoch}')
            return None
        else:
            return self.sched[i]

    def state_dict(self):
        state_dict = [s.state_dict() for s in self.sched]
        return state_dict

    def load_state_dict(self, state, strict=True):
        [self.sched[i].load_state_dict(s) for i, s in enumerate(state)]

    def to(self, device):
        """ Cast schedulers to a certain device.

        Args:
            device (torch.device): Device to cast the scheduler to.
        """
        for sched in self.sched:
            for param in sched.__dict__.values():
                if isinstance(param, torch.Tensor):
                    param.data = param.data.to(device)
                    if param._grad is not None:
                        param._grad.data = param._grad.data.to(device)



#
#   Categorization of graph operations into NodeType's
#   Copyright EAVISE
#
import enum
from .._imports import onnx

__all__ = ['NodeType', 'get_node_type', 'add_nodetype_operation']


class NodeType(enum.Enum):
    """ Enum containing all different types of operation in the dependency graph. """
    CONV = enum.auto()          #: Convolution operation
    BATCHNORM = enum.auto()     #: BatchNorm operation
    CONCAT = enum.auto()        #: Concat operation on channel dimension
    ELEMW_OP = enum.auto()      #: Element-wise operation (eg. add)
    IGNORE = enum.auto()        #: Node that does not need to be modified and does not change output dims (eg. relu)
    IGNORE_STOP = enum.auto()   #: Ignore node and stop dependency chain


def get_node_type(element):
    op_type = element.op_type
    if op_type == 'ATen':
        attr = onnx.helper.printable_attribute(element.attribute[0])
        attr = attr.split('=')[-1].strip()[1:-1]

        for key, value in node_type_map_aten.items():
            if attr in value:
                return key

        raise NotImplementedError(f'ATen: {attr}')
    else:
        for key, value in node_type_map_onnx.items():
            if op_type in value:
                return key

        raise NotImplementedError(element.op_type)


def add_nodetype_operation(name, node_type, aten=True):
    """ Add a graph operation to a specific NodeType. |br|
    Adding all possible ONNX/ATen graph operations to our list is not feasible,
    and thus we test our algorithms only with the models in this library.

    If you get the following error, whilst generating the dependency map of your network,
    it means you use an operation which we have not added to our list of known operations:

    .. code:: bash

       Cannot prune [{layer}], unimplemented dependency [{operation}]

    You can then add your operation to a certain :class:`~lightnet.prune.dependency.NodeType` with this function.
    If the operation name starts with **"ATen:"**,
    add the operation without this prefix and set the ``aten`` argument to **True**.
    Otherwise, set ``aten`` to **False**.

    Args:
        name (str): Name of the operation in the graph (dont forget to remove "ATen:" if necessary)
        node_type (NodeType): To which type to add
        aten (bool, optional): Whether to operation is an ATen or ONNX operation in the graph; Default **True**

    Note:
        We would strongly appreciate it if anyone with unimplemented operations, could open an issue on our gitlab.
        That way, we can manually add it to our list and grow it organically based on usage!
    """
    if aten:
        node_type_map_aten[node_type].add(name)
    else:
        node_type_map_onnx[node_type].add(name)


# Default maps
node_type_map_aten = {
    NodeType.CONV: {
        '_convolution',
    },
    NodeType.BATCHNORM: {
        'batch_norm',
    },
    NodeType.CONCAT: {
        'cat',
    },
    NodeType.ELEMW_OP: {
        'add',
    },
    NodeType.IGNORE: {
        'leaky_relu',
        'relu',
        'hardtanh',
        'adaptive_avg_pool2d',
        'max_pool2d',
        'cummax',
        'upsample_nearest2d',
        'replication_pad2d',
        'flip',
    },
    NodeType.IGNORE_STOP: {
        'size',
    },
}


node_type_map_onnx = {
    NodeType.CONV: set(),
    NodeType.BATCHNORM: set(),
    NodeType.CONCAT: set(),
    NodeType.ELEMW_OP: set(),
    NodeType.IGNORE: set(),
    NodeType.IGNORE_STOP: {
        'Concat',
    },
}



#
#   MobileNet classification network
#   Copyright EAVISE
#

import functools
from collections import OrderedDict
import torch.nn as nn
import lightnet.network as lnn

__all__ = ['MobilenetV2']


class MobilenetV2(lnn.module.Lightnet):
    """ Mobilenet v2 classification network implementation :cite:`mobilenet_v2`.

    Args:
        num_classes (Number, optional): Number of classes; Default **1000**
        alpha (Number, optional): Number between [0-1] that controls the number of filters of the mobilenet convolutions; Default **1**
        input_channels (Number, optional): Number of input channels; Default **3**

    Attributes:
        self.inner_stride: Maximal internal subsampling factor of the network (input dimension should be a multiple of this)

    Note:
        The average pooling is implemented with an :class:`~torch.nn.AdaptiveAvgPool2d` layer. |br|
        For the base input dimension of 224x224, this is exactly the same as a 7x7 average pooling function,
        but the advantage of a adaptive average pooling is that this network can now handle multiple different input dimensions,
        as long as they are a multiple of the ``stride`` factor. |br|
        This is also how the implementation in `tensorflow <mobilenetv2tf_>`_ works.

    Warning:
        When changing the ``alpha`` value, you are changing the network architecture.
        This means you cannot use weights from this network with a different alpha value.

    .. _mobilenetv2tf: https://github.com/tensorflow/models/blob/505f554c6417931c96b59516f14d1ad65df6dbc5/research/slim/nets/mobilenet/mobilenet.py#L365
    """
    inner_stride = 32

    def __init__(self, num_classes, alpha=1, input_channels=3):
        super().__init__()

        # Parameters
        self.num_classes = num_classes
        self.alpha = alpha
        self.input_channels = input_channels

        # Network
        relu = functools.partial(nn.ReLU6, inplace=True)
        self.layers = nn.Sequential(
            # Base layers
            nn.Sequential(OrderedDict([
                ('1_convbatch',     lnn.layer.Conv2dBatchReLU(input_channels, int(alpha*32), 3, 2, 1, relu=relu)),
                ('2_bottleneck',    lnn.layer.InvertedBottleneck(int(alpha*32), int(alpha*16), 3, 1, 1, relu=relu)),
                ('3_bottleneck',    lnn.layer.InvertedBottleneck(int(alpha*16), int(alpha*24), 3, 2, 6, relu=relu)),
                ('4_bottleneck',    lnn.layer.InvertedBottleneck(int(alpha*24), int(alpha*24), 3, 1, 6, relu=relu)),
                ('5_bottleneck',    lnn.layer.InvertedBottleneck(int(alpha*24), int(alpha*32), 3, 2, 6, relu=relu)),
                ('6_bottleneck',    lnn.layer.InvertedBottleneck(int(alpha*32), int(alpha*32), 3, 1, 6, relu=relu)),
                ('7_bottleneck',    lnn.layer.InvertedBottleneck(int(alpha*32), int(alpha*32), 3, 1, 6, relu=relu)),
                ('8_bottleneck',    lnn.layer.InvertedBottleneck(int(alpha*32), int(alpha*64), 3, 2, 6, relu=relu)),
                ('9_bottleneck',    lnn.layer.InvertedBottleneck(int(alpha*64), int(alpha*64), 3, 1, 6, relu=relu)),
                ('10_bottleneck',   lnn.layer.InvertedBottleneck(int(alpha*64), int(alpha*64), 3, 1, 6, relu=relu)),
                ('11_bottleneck',   lnn.layer.InvertedBottleneck(int(alpha*64), int(alpha*64), 3, 1, 6, relu=relu)),
                ('12_bottleneck',   lnn.layer.InvertedBottleneck(int(alpha*64), int(alpha*96), 3, 1, 6, relu=relu)),
                ('13_bottleneck',   lnn.layer.InvertedBottleneck(int(alpha*96), int(alpha*96), 3, 1, 6, relu=relu)),
                ('14_bottleneck',   lnn.layer.InvertedBottleneck(int(alpha*96), int(alpha*96), 3, 1, 6, relu=relu)),
                ('15_bottleneck',   lnn.layer.InvertedBottleneck(int(alpha*96), int(alpha*160), 3, 2, 6, relu=relu)),
                ('16_bottleneck',   lnn.layer.InvertedBottleneck(int(alpha*160), int(alpha*160), 3, 1, 6, relu=relu)),
                ('17_bottleneck',   lnn.layer.InvertedBottleneck(int(alpha*160), int(alpha*160), 3, 1, 6, relu=relu)),
                ('18_bottleneck',   lnn.layer.InvertedBottleneck(int(alpha*160), int(alpha*320), 3, 1, 6, relu=relu)),
                ('19_convbatch',    lnn.layer.Conv2dBatchReLU(int(alpha*320), int(alpha*1280),  1, 1, 0, relu=relu)),
            ])),

            # Classification specific layers
            nn.Sequential(OrderedDict([
                ('20_avgpool',      nn.AdaptiveAvgPool2d(1)),
                ('21_dropout',      nn.Dropout()),
                ('22_conv',         nn.Conv2d(int(alpha*1280), num_classes, 1, 1, 0)),
                ('23_flatten',      lnn.layer.Flatten()),
            ])),
        )



#
#   Lightnet related postprocessing
#   These are functions to transform the output of the network to brambox detection dataframes
#   Copyright EAVISE
#

# Network output to box
from ._cornernet import *
from ._darknet import *

# Util
from ._brambox import *
from ._nms import *
from ._reverse_fit import *



#
#   Lightnet postprocessing for Anchor based detectors (Darknet)
#   Copyright EAVISE
#
import logging
import torch
import torch.nn as nn
from ..util import BaseTransform

__all__ = ['GetCornerBoxes']
log = logging.getLogger(__name__)


class GetCornerBoxes(BaseTransform):
    """ Convert output from cornernet networks to bounding box tensor.

    .. admonition:: Experimental

       This post-processing implementation is still in development
       and might not be yielding the same results as the official implementation.
       Use at your own risk!

    Args:
        embedding_thresh (Number): Embedding distance threshold to filter matching corners
        conf_thresh (Number [0-1]): Confidence threshold to filter detections
        network_stride (Number): Downsampling factor of the network (most lightnet networks have a `inner_stride` attribute)
        topk (Number, optional): Number of corners to select from the network output; Default **100**
        subsample_kernel (Number, optional): Kernel size to perform maxpool subsampling; Default **0**

    Returns:
        (Tensor [Boxes x 7]]): **[batch_num, x_tl, y_tl, x_br, y_br, confidence, class_id]** for every bounding box

    Note:
        If setting the subsample_kernel to **0**, you disable the subsampling.
        Otherwise this post-processing will perform maxpooling on the heatmap with the specified kernel.
    """
    def __init__(self, embedding_thresh, conf_thresh, network_stride, topk=100, subsample_kernel=0):
        super().__init__()
        log.experimental(f'"{self.__class__.__name__}" is still in development. Use at your own risk!')

        self.embedding_thresh = embedding_thresh
        self.conf_thresh = conf_thresh
        self.network_stride = network_stride
        self.topk = topk
        self.subsample_kernel = subsample_kernel

    def forward(self, network_output):
        device = network_output.device
        batch, channels, h, w = network_output.shape

        # Split tensor
        network_output = network_output.view(batch, 2, -1, h, w)        # BATCH, TLBR, NUM_CLASSES+3, H, W
        heatmaps = torch.sigmoid(network_output[:, :, :-3])             # BATCH, TLBR, NUM_CLASSES,   H, W
        embedding = network_output[:, :, -3]                            # BATCH, TLBR,                H, W
        offsets = network_output[:, :, -2:]                             # BATCH, TLBR, XY,            H, W

        # Subsample heatmaps
        if self.subsample_kernel:
            maxpool_heat = nn.functional.max_pool2d(heatmaps.view(batch, -1, h, w), self.subsample_kernel, stride=1, padding=(self.subsample_kernel - 1) // 2)
            heatmaps *= maxpool_heat.view(batch, 2, -1, h, w) == heatmaps

        # Get topK corners
        topk_heatmaps, topk_idx = torch.topk(heatmaps.view(batch, 2, -1), self.topk)
        topk_classes = topk_idx // (h * w)
        topk_idx %= (h * w)
        topk_x = (topk_idx % w).float()
        topk_y = (topk_idx // w).float()

        # Add XY offsets
        offset_x = torch.gather(offsets[:, :, 0].reshape(batch, 2, -1), 2, topk_idx)
        offset_y = torch.gather(offsets[:, :, 1].reshape(batch, 2, -1), 2, topk_idx)
        topk_x = topk_x + offset_x
        topk_y = topk_y + offset_y

        # Combine TL and BR corners
        tl_x = topk_x[:, 0, :, None].expand(-1, self.topk, self.topk)
        tl_y = topk_y[:, 0, :, None].expand(-1, self.topk, self.topk)
        br_x = topk_x[:, 1, None, :].expand(-1, self.topk, self.topk)
        br_y = topk_y[:, 1, None, :].expand(-1, self.topk, self.topk)
        bboxes = torch.stack([tl_x, tl_y, br_x, br_y], dim=3)
        bboxes *= self.network_stride

        # Create corner filter
        corner_filter = (br_x >= tl_x) & (br_y >= tl_y)

        # Create class filter
        tl_classes = topk_classes[:, 0, :, None].expand(-1, self.topk, self.topk)
        br_classes = topk_classes[:, 1, None, :].expand(-1, self.topk, self.topk)
        class_filter = (tl_classes == br_classes)

        # Create confidence filter
        # NOTE : This is different than the original implementation, where they keep the TOP N detections
        confidence = (topk_heatmaps[:, 0, :, None] + topk_heatmaps[:, 1, None, :]) / 2
        confidence_filter = confidence > self.conf_thresh

        # Create embedding filter
        topk_embed = torch.gather(embedding.view(batch, 2, -1), 2, topk_idx)
        dist = torch.abs(topk_embed[:, 0, :, None] - topk_embed[:, 1, None, :])
        embedding_filter = dist <= self.embedding_thresh

        # Get batch number of the detections
        total_filter = class_filter & embedding_filter & corner_filter & confidence_filter
        nums = torch.arange(0, batch, dtype=torch.uint8, device=total_filter.device)
        batch_num = total_filter.view(batch, -1)
        batch_num = nums[:, None].expand_as(batch_num)[batch_num]

        # Apply filters and combine values
        bboxes = bboxes[total_filter, :].view(-1, 4)
        confidence = confidence[total_filter].view(-1)
        class_idx = tl_classes[total_filter].view(-1)

        return torch.cat([batch_num[:, None].float(), bboxes, confidence[:, None], class_idx[:, None].float()], dim=1)



#
#   Lightnet related postprocessing
#   Copyright EAVISE
#

import logging
import numpy as np
import torch
from ..util import BaseTransform
from ..._imports import bb


__all__ = ['NMS', 'NMSFast', 'NMSSoft', 'NMSSoftFast', 'NonMaxSuppression']
log = logging.getLogger(__name__)


class NMS(BaseTransform):
    """ Performs non-maximal suppression on the bounding boxes, filtering boxes with a high overlap.
    This function can either work on a pytorch bounding box tensor or a brambox dataframe.

    Args:
        nms_thresh (Number [0-1]): Overlapping threshold to filter detections with non-maxima suppresion
        class_nms (Boolean, optional): Whether to perform nms per class; Default **True**
        force_cpu (Boolean, optional): Whether to force a part of the computation on CPU (tensor only, see Note); Default **True**
        reset_index (Boolean, optional): Whether to reset the index of the returned dataframe (dataframe only); Default **True**

    Input:
        boxes (Tensor [Boxes x 7] or pandas.Dataframe): bounding boxes

    Returns:
        boxes (Tensor [Boxes x 7] or pandas.Dataframe): filtered bounding boxes

    Note:
        This post-processing function expects the input bounding boxes to be either a PyTorch tensor or a brambox dataframe.

        The PyTorch tensor needs to be formatted as follows: **[batch_num, x_tl, y_tl, x_br, y_br, confidence, class_id]** for every bounding box.
        This corresponds to the output from the Get***Boxes classes available in lightnet.

        The brambox dataframe should be a detection dataframe,
        as it needs the x_top_left, y_top_left, width, height, confidence and class_label columns.

    Note:
        A part of the computation of NMS involve a sequential loop through the boxes.
        This is quite difficult to implement efficiently on the GPU, using the PyTorch API.
        We thus concluded empirically that it is usually more efficient to move GPU tensors to the CPU for this part of the computations. |br|
        By passing false to `force_cpu`, you can disable this behavior and perform all the computations on the original device of the input tensor.
    """
    def __init__(self, nms_thresh, class_nms=True, force_cpu=True, reset_index=True):
        super().__init__()
        self.nms_thresh = nms_thresh
        self.class_nms = class_nms
        self.force_cpu = force_cpu
        self.reset_index = reset_index

    def forward(self, boxes):
        if isinstance(boxes, torch.Tensor):
            return self._torch(boxes)
        else:
            return self._pandas(boxes)

    def _torch(self, boxes):
        if boxes.numel() == 0:
            return boxes

        batches = boxes[:, 0]
        keep = torch.empty(boxes.shape[0], dtype=torch.bool, device=boxes.device)
        for batch in torch.unique(batches, sorted=False):
            mask = batches == batch
            keep[mask] = self._torch_nms(boxes[mask])

        return boxes[keep]

    def _torch_nms(self, boxes):
        if boxes.numel() == 0:
            return boxes

        bboxes = boxes[:, 1:5]
        scores = boxes[:, 5]
        classes = boxes[:, 6]

        # Sort coordinates by descending score
        scores, order = scores.sort(0, descending=True)
        x1, y1, x2, y2 = bboxes[order].split(1, 1)

        # Compute dx and dy between each pair of boxes (these mat contain every pair twice...)
        dx = (x2.min(x2.t()) - x1.max(x1.t())).clamp(min=0)
        dy = (y2.min(y2.t()) - y1.max(y1.t())).clamp(min=0)

        # Compute iou
        intersections = dx * dy
        areas = (x2 - x1) * (y2 - y1)
        unions = (areas + areas.t()) - intersections
        ious = intersections / unions

        # Filter based on iou (and class)
        conflicting = (ious > self.nms_thresh).triu(1)
        if self.class_nms:
            classes = classes[order]
            same_class = (classes.unsqueeze(0) == classes.unsqueeze(1))
            conflicting = (conflicting & same_class)

        if self.force_cpu:
            conflicting = conflicting.cpu()
        keep = torch.zeros(conflicting.shape[0], dtype=torch.bool, device=conflicting.device)
        supress = torch.zeros(conflicting.shape[0], dtype=torch.bool, device=conflicting.device)
        for i, row in enumerate(conflicting):
            if not supress[i]:
                keep[i] = True
                supress[row] = True

        keep = keep.to(boxes.device)
        return keep.scatter(0, order, keep)

    @torch.jit.ignore
    def _pandas(self, boxes):
        if len(boxes.index) == 0:
            return boxes
        boxes = boxes.groupby('image', group_keys=False, observed=True).apply(self._pandas_nms)
        if self.reset_index:
            return boxes.reset_index(drop=True)
        return boxes

    def _pandas_nms(self, boxes):
        boxes = boxes.sort_values('confidence', ascending=False)
        ious = bb.stat.coordinates.iou(boxes, boxes, bias=0)
        scores = boxes['confidence'].values

        # Filter based on iou (and class)
        conflicting = np.triu(ious > self.nms_thresh, 1)
        if self.class_nms:
            classes = boxes['class_label'].values
            same_class = (classes[None, ...] == classes[..., None])
            conflicting = (conflicting & same_class)

        keep = np.zeros(conflicting.shape[0], dtype=np.bool)
        supress = np.zeros(conflicting.shape[0], dtype=np.bool)
        for i, row in enumerate(conflicting):
            if not supress[i]:
                keep[i] = True
                supress[row] = True

        return boxes[keep]


class NMSFast(NMS):
    """ Performs fast NMS on the bounding boxes, filtering boxes with a high overlap.
    This function can either work on a pytorch bounding box tensor or a brambox dataframe.

    This faster alternative makes a small "mistake" during NMS computation,
    in order to remove a necessary loop in the code, allowing it to run faster.
    The speed increase should be mostly notable when performing NMS with PyTorch tensors on the GPU.

    The difference is explained in the image below, where the boxes A and B overlap enough to be filtered out
    and the boxes B and C as well (but A and C do not). |br|
    Regular NMS will keep box C in this situation, because box B gets filtered out and is thus not there to remove C.
    Fast NMS will not do this and will only keep box A in this situation. |br|
    Depending on the use-case (closely clustered and overlapping objects), this might be a problem or not.

    .. figure:: /.static/api/nms-fast.*
       :width: 100%
       :alt: Fast NMS problem

       Regular NMS will keep both boxes A and C, but Fast NMS will only keep A in this example.

    Args:
        nms_thresh (Number [0-1]): Overlapping threshold to filter detections with non-maxima suppresion
        class_nms (Boolean, optional): Whether to perform nms per class; Default **True**
        reset_index (Boolean, optional): Whether to reset the index of the returned dataframe (dataframe only); Default **True**

    Input:
        boxes (Tensor [Boxes x 7] or pandas.Dataframe): bounding boxes

    Returns:
        boxes (Tensor [Boxes x 7] or pandas.Dataframe): filtered bounding boxes

    Note:
        This post-processing function expects the input bounding boxes to be either a PyTorch tensor or a brambox dataframe.

        The PyTorch tensor needs to be formatted as follows: **[batch_num, x_tl, y_tl, x_br, y_br, confidence, class_id]** for every bounding box.
        This corresponds to the output from the Get***Boxes classes available in lightnet.

        The brambox dataframe should be a detection dataframe,
        as it needs the x_top_left, y_top_left, width, height, confidence and class_label columns.
    """
    def __init__(self, nms_thresh, class_nms=True, reset_index=True):
        super().__init__(nms_thresh, class_nms, False, reset_index)

    def _torch_nms(self, boxes):
        if boxes.numel() == 0:
            return boxes

        bboxes = boxes[:, 1:5]
        scores = boxes[:, 5]
        classes = boxes[:, 6]

        # Sort coordinates by descending score
        scores, order = scores.sort(0, descending=True)
        x1, y1, x2, y2 = bboxes[order].split(1, 1)

        # Compute dx and dy between each pair of boxes (these mat contain every pair twice...)
        dx = (x2.min(x2.t()) - x1.max(x1.t())).clamp(min=0)
        dy = (y2.min(y2.t()) - y1.max(y1.t())).clamp(min=0)

        # Compute iou
        intersections = dx * dy
        areas = (x2 - x1) * (y2 - y1)
        unions = (areas + areas.t()) - intersections
        ious = intersections / unions

        # Filter based on iou (and class)
        conflicting = (ious > self.nms_thresh).triu(1)
        if self.class_nms:
            classes = classes[order]
            same_class = (classes.unsqueeze(0) == classes.unsqueeze(1))
            conflicting = (conflicting & same_class)

        keep = conflicting.sum(0) == 0
        return keep.scatter(0, order, keep)

    @torch.jit.ignore
    def _pandas_nms(self, boxes):
        boxes = boxes.sort_values('confidence', ascending=False)
        ious = bb.stat.coordinates.iou(boxes, boxes, bias=0)
        scores = boxes['confidence'].values

        conflicting = np.triu(ious > self.nms_thresh, 1)
        if self.class_nms:
            classes = boxes['class_label'].values
            same_class = (classes[None, ...] == classes[..., None])
            conflicting = (conflicting & same_class)

        # Return filtered boxes
        keep = conflicting.sum(0) == 0
        return boxes[keep]


class NMSSoft(BaseTransform):
    """ Performs soft NMS with exponential decaying on the bounding boxes, as explained in :cite:`soft_nms`.
    This function can either work on a pytorch bounding box tensor or a brambox dataframe.

    Args:
        sigma (Number): Sensitivity value for the confidence rescaling (exponential decay)
        conf_thresh (Number [0-1], optional): Confidence threshold to filter the bounding boxes after decaying them; Default **0**
        class_nms (Boolean, optional): Whether to perform nms per class; Default **True**
        force_cpu (Boolean, optional): Whether to force a part of the computation on CPU (tensor only, see Note); Default **True**
        reset_index (Boolean, optional): Whether to reset the index of the returned dataframe (dataframe only); Default **True**

    Input:
        boxes (Tensor [Boxes x 7] or pandas.Dataframe): bounding boxes

    Returns:
        boxes (Tensor [Boxes x 7] or pandas.Dataframe): filtered bounding boxes

    Note:
        This post-processing function expects the input bounding boxes to be either a PyTorch tensor or a brambox dataframe.

        The PyTorch tensor needs to be formatted as follows: **[batch_num, x_tl, y_tl, x_br, y_br, confidence, class_id]** for every bounding box.
        This corresponds to the output from the Get***Boxes classes available in lightnet.

        The brambox dataframe should be a detection dataframe,
        as it needs the x_top_left, y_top_left, width, height, confidence and class_label columns.
    """
    def __init__(self, sigma, conf_thresh=0, class_nms=True, force_cpu=True, reset_index=True):
        super().__init__()
        self.sigma = sigma
        self.conf_thresh = conf_thresh
        self.class_nms = class_nms
        self.force_cpu = force_cpu
        self.reset_index = reset_index

    def forward(self, boxes):
        if isinstance(boxes, torch.Tensor):
            return self._torch(boxes)
        else:
            return self._pandas(boxes)

    def _torch(self, boxes):
        if boxes.numel() == 0:
            return boxes

        batches = boxes[:, 0]
        for batch in torch.unique(batches, sorted=False):
            mask = batches == batch
            boxes[mask, 5] = self._torch_nms(boxes[mask])

        if self.conf_thresh > 0:
            keep = boxes[:, 5] > self.conf_thresh
            return boxes[keep]

        return boxes

    def _torch_nms(self, boxes):
        if boxes.numel() == 0:
            return boxes

        bboxes = boxes[:, 1:5]
        scores = boxes[:, 5]
        classes = boxes[:, 6]

        # Sort coordinates by descending score
        scores, order = scores.sort(0, descending=True)
        x1, y1, x2, y2 = bboxes[order].split(1, 1)

        # Compute dx and dy between each pair of boxes (these mat contain every pair twice...)
        dx = (x2.min(x2.t()) - x1.max(x1.t())).clamp(min=0)
        dy = (y2.min(y2.t()) - y1.max(y1.t())).clamp(min=0)

        # Compute iou
        intersections = dx * dy
        areas = (x2 - x1) * (y2 - y1)
        unions = (areas + areas.t()) - intersections
        ious = intersections / unions

        # Filter class
        if self.class_nms:
            classes = classes[order]
            same_class = (classes.unsqueeze(0) == classes.unsqueeze(1))
            ious *= same_class

        # Decay scores
        decay = torch.exp(-(ious ** 2) / self.sigma)
        if self.force_cpu:
            scores = scores.cpu()
            order = order.cpu()
            decay = decay.cpu()

        tempscores = scores.clone()
        for _ in range(scores.shape[0]):
            maxidx = tempscores.argmax()
            maxscore = tempscores[maxidx]
            if maxscore <= self.conf_thresh:
                break

            tempscores[maxidx] = -1
            mask = tempscores != -1
            tempscores[mask] *= decay[maxidx, mask]
            scores[mask] = tempscores[mask]

        scores = scores.to(boxes.device)
        order = order.to(boxes.device)
        return scores.scatter(0, order, scores)

    @torch.jit.ignore
    def _pandas(self, boxes):
        if len(boxes.index) == 0:
            return boxes

        boxes = boxes.groupby('image', group_keys=False, observed=True).apply(self._pandas_nms)
        if self.conf_thresh > 0:
            boxes = boxes[boxes.confidence > self.conf_thresh].copy()
        if self.reset_index:
            return boxes.reset_index(drop=True)

        return boxes

    def _pandas_nms(self, boxes):
        boxes = boxes.sort_values('confidence', ascending=False)
        scores = boxes['confidence'].values
        ious = bb.stat.coordinates.iou(boxes, boxes, bias=0)

        # Filter class
        if self.class_nms:
            classes = boxes['class_label'].values
            same_class = (classes[None, ...] == classes[..., None])
            ious *= same_class

        # Decay scores
        decay = np.exp(-(ious ** 2) / self.sigma)
        tempscores = scores.copy()
        for _ in range(scores.shape[0]):
            maxidx = tempscores.argmax()
            maxscore = tempscores[maxidx]
            if maxscore <= self.conf_thresh:
                break

            tempscores[maxidx] = -1
            mask = tempscores != -1
            tempscores[mask] *= decay[maxidx, mask]
            scores[mask] = tempscores[mask]

        # Set scores back
        boxes['confidence'] = scores
        return boxes


class NMSSoftFast(BaseTransform):
    """ Faster version of SoftNMS which filters boxes with a high overlap, using exponential decay.
    This function can either work on a pytorch bounding box tensor or a brambox dataframe.

    This version of NMS makes the same "mistake" as :class:`~lightnet.data.transform.NMSFast`,
    which in turn allows it to be faster than the regular :class:`~lightnet.data.transform.NMSSoft` algorithm.
    The speed increase should be mostly notable when performing NMS with PyTorch tensors on the GPU.

    Args:
        sigma (Number): Sensitivity value for the confidence rescaling (exponential decay)
        conf_thresh (Number [0-1], optional): Confidence threshold to filter the bounding boxes after decaying them; Default **0**
        class_nms (Boolean, optional): Whether to perform nms per class; Default **True**
        reset_index (Boolean, optional): Whether to reset the index of the returned dataframe (dataframe only); Default **True**

    Input:
        boxes (Tensor [Boxes x 7] or pandas.Dataframe): bounding boxes

    Returns:
        boxes (Tensor [Boxes x 7] or pandas.Dataframe): filtered bounding boxes

    Note:
        This post-processing function expects the input bounding boxes to be either a PyTorch tensor or a brambox dataframe.

        The PyTorch tensor needs to be formatted as follows: **[batch_num, x_tl, y_tl, x_br, y_br, confidence, class_id]** for every bounding box.
        This corresponds to the output from the Get***Boxes classes available in lightnet.

        The brambox dataframe should be a detection dataframe,
        as it needs the x_top_left, y_top_left, width, height, confidence and class_label columns.
    """
    def __init__(self, sigma, conf_thresh=0, class_nms=True, reset_index=True):
        super().__init__()
        self.sigma = sigma
        self.conf_thresh = conf_thresh
        self.class_nms = class_nms
        self.reset_index = reset_index

    def forward(self, boxes):
        if isinstance(boxes, torch.Tensor):
            return self._torch(boxes)
        else:
            return self._pandas(boxes)

    def _torch(self, boxes):
        if boxes.numel() == 0:
            return boxes

        batches = boxes[:, 0]
        for batch in torch.unique(batches, sorted=False):
            mask = batches == batch
            boxes[mask, 5] = self._torch_nms(boxes[mask])

        if self.conf_thresh > 0:
            keep = boxes[:, 5] > self.conf_thresh
            return boxes[keep]

        return boxes

    def _torch_nms(self, boxes):
        if boxes.numel() == 0:
            return boxes

        bboxes = boxes[:, 1:5]
        scores = boxes[:, 5]
        classes = boxes[:, 6]

        # Sort coordinates by descending score
        scores, order = scores.sort(0, descending=True)
        x1, y1, x2, y2 = bboxes[order].split(1, 1)

        # Compute dx and dy between each pair of boxes (these mat contain every pair twice...)
        dx = (x2.min(x2.t()) - x1.max(x1.t())).clamp(min=0)
        dy = (y2.min(y2.t()) - y1.max(y1.t())).clamp(min=0)

        # Compute iou
        intersections = dx * dy
        areas = (x2 - x1) * (y2 - y1)
        unions = (areas + areas.t()) - intersections
        ious = intersections / unions

        # Filter class
        if self.class_nms:
            classes = classes[order]
            same_class = (classes.unsqueeze(0) == classes.unsqueeze(1))
            ious *= same_class

        # Decay scores
        decay = ious.triu(1)
        decay = torch.exp(-(decay ** 2) / self.sigma).prod(0)
        scores *= decay
        return scores.scatter(0, order, scores)

    @torch.jit.ignore
    def _pandas(self, boxes):
        if len(boxes.index) == 0:
            return boxes

        boxes = boxes.groupby('image', group_keys=False, observed=True).apply(self._pandas_nms)
        if self.conf_thresh > 0:
            boxes = boxes[boxes.confidence > self.conf_thresh].copy()
        if self.reset_index:
            return boxes.reset_index(drop=True)

        return boxes

    def _pandas_nms(self, boxes):
        boxes = boxes.sort_values('confidence', ascending=False)
        scores = boxes['confidence'].values
        ious = bb.stat.coordinates.iou(boxes, boxes, bias=0)

        # Filter class
        if self.class_nms:
            classes = boxes['class_label'].values
            same_class = (classes[None, ...] == classes[..., None])
            ious *= same_class

        # Decay scores
        decay = np.triu(ious, 1)
        decay = np.prod(np.exp(-(decay ** 2) / self.sigma), 0)
        scores *= decay

        # Set scores back
        boxes['confidence'] = scores
        return boxes


def NonMaxSuppression(*args, **kwargs):
    log.deprecated('NonMaxSuppression is deprecated, please use "NMS"')
    return NMS(*args, **kwargs)



#
#   Base lightnet network module structure
#   Copyright EAVISE
#

import inspect
import logging
import re
from collections import OrderedDict
import torch
import torch.nn as nn

__all__ = ['Lightnet']
log = logging.getLogger(__name__)


class Lightnet(nn.Module):
    """ This class provides an abstraction layer on top of :class:`pytorch:torch.nn.Module` and is used as a base for every network implemented in this framework.

    Note:
        If you define **self.layers** as a :class:`pytorch:torch.nn.Sequential` or :class:`pytorch:torch.nn.ModuleList`,
        the default ``forward()`` function can use these layers automatically to run the network.
    """
    def __init__(self):
        super().__init__()
        self.layers = None

    def forward(self, x):
        log.debug('Running default forward function')
        if hasattr(self, 'layers'):
            if isinstance(self.layers, nn.Sequential):
                return self.layers(x)
            elif isinstance(self.layers, nn.ModuleList):
                log.warning('No _forward function defined, looping sequentially over modulelist')
                for _, module in enumerate(self.layers):
                    x = module(x)
                return x
            else:
                raise NotImplementedError(f'No _forward function defined and no default behaviour for this type of layers [{type(self.layers)}]')
        else:
            raise NotImplementedError(f'No _forward function defined and no default behaviour for this network')

    def layer_loop(self, mod=None):
        """ This function will recursively loop over all moduleList and Sequential children.

        Args:
            mod (torch.nn.Module, optional): Module to loop over; Default **self**

        Returns:
            (generator): Iterator that will loop over and yield the different layers
        """
        if mod is None:
            mod = self

        for module in mod.children():
            if isinstance(module, (nn.ModuleList, nn.Sequential)):
                yield from self.layer_loop(module)
            else:
                yield module

    def named_layer_loop(self, mod=None):
        """ Named version of :func:`~lightnet.network.module.Lightnet.layer_loop`

        Args:
            mod (torch.nn.Module, optional): Module to loop over; Default **self**

        Returns:
            (generator): Iterator that will loop over and yield (name, layer) tuples
        """
        if mod is None:
            mod = self

        for name, module in mod.named_children():
            if isinstance(module, (nn.ModuleList, nn.Sequential)):
                yield from self.named_layer_loop(module)
            else:
                yield name, module

    def load(self, weights_file, remap=None, strict=True):
        """ This function will load the weights from a file.
        It also allows to load in a weights file with only a part of the weights in.

        Args:
            weights_file (str): path to file
            remap (callable or list, optional): Remapping of the weights, see :func:`~lightnet.network.module.Lightnet.weight_remapping`; Default **None**
            strict (Boolean, optional): Whether the weight file should contain all layers of the model; Default **True**

        Note:
            This function will load the weights to CPU,
            so you should use ``network.to(device)`` afterwards to send it to the device of your choice.
        """
        keys = self.state_dict().keys()

        if remap is not None:
            state = self.weight_remapping(torch.load(weights_file, 'cpu'), remap)
            remap = ' remapped'
        else:
            state = torch.load(weights_file, 'cpu')
            remap = ''

        log.info(f'Loading{remap} weights from file [{weights_file}]')
        if not strict and state.keys() != keys:
            log.warning('Modules not matching, performing partial update')

        self.load_state_dict(state, strict=strict)

    def load_pruned(self, weights_file, strict=True):
        """ This function will load pruned weights from a file.
        It also allows to load a weights file,
        which contains less channels in a convolution than orginally defined in the network.

        Args:
            weights_file (str): path to file
            strict (Boolean, optional): Whether the weight file should contain all layers of the model; Default **True**

        Note:
            This function will load the weights to CPU,
            so you should use ``network.to(device)`` afterwards to send it to the device of your choice.
        """
        keys = set(self.state_dict().keys())
        log.info(f'Loading pruned weights from file [{weights_file}]')
        state = torch.load(weights_file, 'cpu')

        # Prune tensors
        for key, val in state.items():
            if key in keys:
                tensor = self
                module = self
                name = None
                for p in key.split("."):
                    module = tensor
                    name = p
                    tensor = getattr(tensor, p)

                if tensor.shape != val.shape:
                    slices = [slice(0, s) for s in val.shape]
                    if isinstance(tensor, torch.nn.Parameter):
                        setattr(module, p, torch.nn.Parameter(tensor[slices]))
                    else:
                        setattr(module, p, tensor[slices])

        # Modify module metadata
        for module in self.modules():
            if isinstance(module, torch.nn.Conv2d):
                if module.groups == 1:
                    module.in_channels = module.weight.shape[1]
                    module.out_channels = module.weight.shape[0]
                elif module.groups == module.in_channels == module.out_channels:
                    module.out_channels = module.weight.shape[0]
                    module.in_channels = module.out_channels
                    module.groups = module.out_channels
            elif isinstance(module, torch.nn.BatchNorm2d):
                if module.weight is not None:
                    module.num_features = module.weight.shape[0]
                elif module.running_mean is not None:
                    module.num_features = module.running_mean.shape[0]

        # Load weights
        if not strict and state.keys() != keys:
            log.warning('Modules not matching, performing partial update')
        self.load_state_dict(state, strict=strict)

    def save(self, weights_file, remap=None):
        """ This function will save the weights to a file.

        Args:
            weights_file (str): path to file
            remap (callable or list, optional): Remapping of the weights, see :func:`~lightnet.network.module.Lightnet.weight_remapping`; Default **None**
        """
        if remap is not None:
            state = self.weight_remapping(self.state_dict(), remap)
            remap = ' remapped'
        else:
            state = self.state_dict()
            remap = ''

        torch.save(state, weights_file)
        log.info(f'Saved{remap} weights as {weights_file}')

    def __str__(self):
        """ Shorter version than default PyTorch one. """
        args = list(inspect.signature(self.__class__.__init__).parameters.keys())
        args.remove('self')

        string = self.__class__.__module__ + '.' + self.__class__.__name__ + '('
        for i, arg in enumerate(args):
            if i > 0:
                string += ', '
            val = getattr(self, arg, '?')
            string += f'{arg}={val}'
        string += ')'

        return string

    @staticmethod
    def weight_remapping(weights, remap):
        """ This function is used to remap the keys of a ``state_dict``.
        This can be usefull to load in weights from a different framework, or to modify weights from a backbone network, for usage in another (detection) network. |br|
        This method does not usually get called directly, but is used by :func:`~lightnet.network.module.Lightnet.load` and :func:`~lightnet.network.module.Lightnet.save`
        to modify the weights prior to loading/saving them.

        Args:
            weights (dict): The weights state dictionary
            remap (callable or list): Remapping of the weights, see Note

        Note:
            The optional ``remap`` parameter expects a callable object or a list of tuples.

            If the argument is callable, it will be called with each key in the ``state_dict`` and it should return a new key.
            If it returns **None**, that weight will be removed from the new state_dict.

            if the ``remap`` argument is a list of tuples, they should contain **('old', 'new')** remapping sequences.
            The remapping sequence can contain strings or regex objects. |br|
            What happens when you supply a remapping list,
            is that this function will loop over the ``state_dict`` of the model and for each parameter of the ``state_dict`` it will loop through the remapping list.
            If the first string or regex of the remapping sequence is found in the ``state_dict`` key, it will be replaced by the second string or regex of that remapping sequence. |br|
            There are two important things to note here:

            - If a key does not match any remapping sequence, it gets discarded. To save all the weights, even if you need no remapping, add a last remapping sequence of **('', '')** which will match with all keys, but not modify them.
            - The remapping sequences or processed in order. This means that if a key matches with a certain remapping sequence, the following sequences will not be considered anymore.
        """
        new_weights = OrderedDict()

        if callable(remap):
            for k, v in weights.items():
                nk = remap(k)
                if nk is not None:
                    new_weights[nk] = v
        else:
            for k, v in weights.items():
                for r in remap:
                    if re.match(r[0], k) is not None:
                        new_weights[re.sub(r[0], r[1], k)] = v
                        break

        return new_weights



#
#   Lightnet dataset and dataloading mechanisms
#   Copyright EAVISE
#

import random
import logging
import collections
from functools import wraps
import torch
from torch.utils.data.dataset import Dataset as torchDataset
from torch.utils.data.sampler import BatchSampler as torchBatchSampler
from torch.utils.data.dataloader import DataLoader as torchDataLoader
from torch.utils.data.dataloader import default_collate
from ._imports import pd, bb


__all__ = ['Dataset', 'DataLoader', 'brambox_collate']
log = logging.getLogger(__name__)


class Dataset(torchDataset):
    """ This class is a subclass of the base :class:`torch.utils.data.Dataset`,
    that enables on the fly resizing of the ``input_dim`` with a :class:`lightnet.data.DataLoader`.

    Args:
        input_dimension (tuple, optional): (width,height) tuple with default dimensions of the network; Default **[None, None]**
    """
    def __init__(self, input_dimension=None):
        super().__init__()
        self.__input_dim = input_dimension[:2] if input_dimension is not None else [None, None]

    @property
    def input_dim(self):
        """ Dimension that can be used by transforms to set the correct image size, etc.
        This allows transforms to have a single source of truth for the input dimension of the network.

        Return:
            list: Tuple containing the current width,height
        """
        if hasattr(self, '_input_dim'):
            return self._input_dim
        return self.__input_dim

    @staticmethod
    def resize_getitem(getitem_fn):
        """ Decorator method that needs to be used around the ``__getitem__`` method. |br|
        This decorator enables the on the fly resizing  of the ``input_dim`` with our :class:`~lightnet.data.DataLoader` class.

        Example:
            >>> class CustomSet(ln.data.Dataset):
            ...     def __len__(self):
            ...         return 10
            ...     @ln.data.Dataset.resize_getitem
            ...     def __getitem__(self, index):
            ...         # Should return (image, anno) but here we return input_dim
            ...         return self.input_dim
            >>> data = CustomSet((200,200))
            >>> data[0]
            (200, 200)
            >>> data[(480,320), 0]
            (480, 320)
        """
        @wraps(getitem_fn)
        def wrapper(self, index):
            if not isinstance(index, int):
                has_dim = True
                self._input_dim = index[0]
                index = index[1]
            else:
                has_dim = False

            ret_val = getitem_fn(self, index)

            if has_dim:
                del self._input_dim

            return ret_val

        return wrapper


class DataLoader(torchDataLoader):
    """ Lightnet dataloader that enables on the fly resizing of the images.
    See :class:`torch.utils.data.DataLoader` for more information on the arguments.

    Note:
        This dataloader only works with :class:`lightnet.data.Dataset` based datasets.

    Example:
        >>> class CustomSet(ln.data.Dataset):
        ...     def __len__(self):
        ...         return 4
        ...     @ln.data.Dataset.resize_getitem
        ...     def __getitem__(self, index):
        ...         # Should return (image, anno) but here we return (input_dim,)
        ...         return (self.input_dim,)
        >>> dl = ln.data.DataLoader(
        ...     CustomSet((200,200)),
        ...     batch_size = 2,
        ... )
        >>> dl.dataset.input_dim    # Default input_dim
        (200, 200)
        >>> for d in dl:
        ...     d
        [[tensor([200, 200]), tensor([200, 200])]]
        [[tensor([200, 200]), tensor([200, 200])]]
        >>> dl.change_input_dim(320, random_range=None)
        (320, 320)
        >>> for d in dl:
        ...     d
        [[tensor([320, 320]), tensor([320, 320])]]
        [[tensor([320, 320]), tensor([320, 320])]]
        >>> dl.change_input_dim((480, 320), random_range=None)
        (480, 320)
        >>> for d in dl:
        ...     d
        [[tensor([480, 480]), tensor([320, 320])]]
        [[tensor([480, 480]), tensor([320, 320])]]
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__initialized = False
        shuffle = False
        sampler = None
        batch_sampler = None
        if len(args) > 5:
            shuffle = args[2]
            sampler = args[3]
            batch_sampler = args[4]
        elif len(args) > 4:
            shuffle = args[2]
            sampler = args[3]
            if 'batch_sampler' in kwargs:
                batch_sampler = kwargs['batch_sampler']
        elif len(args) > 3:
            shuffle = args[2]
            if 'sampler' in kwargs:
                sampler = kwargs['sampler']
            if 'batch_sampler' in kwargs:
                batch_sampler = kwargs['batch_sampler']
        else:
            if 'shuffle' in kwargs:
                shuffle = kwargs['shuffle']
            if 'sampler' in kwargs:
                sampler = kwargs['sampler']
            if 'batch_sampler' in kwargs:
                batch_sampler = kwargs['batch_sampler']

        # Use custom BatchSampler
        if batch_sampler is None:
            if sampler is None:
                if shuffle:
                    sampler = torch.utils.data.sampler.RandomSampler(self.dataset)
                else:
                    sampler = torch.utils.data.sampler.SequentialSampler(self.dataset)
            batch_sampler = BatchSampler(sampler, self.batch_size, self.drop_last, input_dimension=self.dataset.input_dim)

        self.sampler = sampler
        self.batch_sampler = batch_sampler

        self.__initialized = True

    def change_input_dim(self, multiple=32, random_range=(10, 19)):
        """ This function will compute a new size and update it on the next mini_batch.

        Args:
            multiple (int or tuple, optional): value (or values) to multiply the randomly generated range by; Default **32**
            random_range (tuple, optional): This (min, max) tuple sets the range for the randomisation; Default **(10, 19)**

        Return:
            tuple: width, height tuple with new dimension

        Note:
            The new size is generated as follows: |br|
            First we compute a random integer inside ``[random_range]``.
            We then multiply that number with the ``multiple`` argument, which gives our final new input size. |br|
            If ``multiple`` is an integer we generate a square size. If you give a tuple of **(width, height)**,
            the size is computed as :math:`rng * multiple[0], rng * multiple[1]`.

        Note:
            You can set the ``random_range`` argument to **None** to set an exact size of multiply. |br|
            See the example above for how this works.
        """
        if random_range is None:
            size = 1
        else:
            size = random.randint(*random_range)

        if isinstance(multiple, int):
            size = (size * multiple, size * multiple)
        else:
            size = (size * multiple[0], size * multiple[1])

        self.batch_sampler.new_input_dim = size

        return size


class BatchSampler(torchBatchSampler):
    """ This batch sampler will generate mini-batches of (dim, index) tuples from another sampler.
    It works just like the :class:`torch.utils.data.sampler.BatchSampler`, but it will prepend a dimension,
    whilst ensuring it stays the same across one mini-batch.
    """
    def __init__(self, *args, input_dimension=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.input_dim = input_dimension
        self.new_input_dim = None

    def __iter__(self):
        self.__set_input_dim()
        for batch in super().__iter__():
            yield [(self.input_dim, idx) for idx in batch]
            self.__set_input_dim()

    def __set_input_dim(self):
        """ This function randomly changes the the input dimension of the dataset. """
        if self.new_input_dim is not None:
            log.info(f'Resizing network {self.new_input_dim[:2]}')
            self.input_dim = (self.new_input_dim[0], self.new_input_dim[1])
            self.new_input_dim = None


def brambox_collate(batch):
    """ Function that collates dataframes by concatenating them.

    Note:
        If the dataframes contain an 'image' categorical column (aka. brambox dataframes),
        they will be concatenated with the :func:`brambox.util.concat` function.
    """
    if isinstance(batch[0], pd.DataFrame):
        for i, df in enumerate(batch):
            df['batch_number'] = i
        if 'image' in batch[0].columns and batch[0].image.dtype == 'category':
            return bb.util.concat(batch, ignore_index=True, sort=False)
        else:
            return pd.concat(batch, ignore_index=True, sort=False)
    elif isinstance(batch[0], collections.abc.Sequence) and not isinstance(batch[0], (str, bytes)):
        transposed = zip(*batch)
        return [brambox_collate(samples) for samples in transposed]
    else:
        return default_collate(batch)



"""
Lightnet Models Module |br|
This module contains networks that were recreated with this library.
Take a look at the code to learn how to use this library, or just use these models if that is all you need.
"""

# Lightnet
from ._dataset_brambox import *

# Darknet
from ._dataset_darknet import *
from ._network_darknet import *
from ._network_darknet19 import *
from ._network_darknet53 import *
from ._network_tiny_yolo_v2 import *
from ._network_tiny_yolo_v3 import *
from ._network_yolo_v2 import *
from ._network_yolo_v3 import *

# Yolt
from ._network_yolt import *

# DYolo
from ._network_dyolo import *

# Mobilenet
from ._network_mobilenet_v1 import *
from ._network_mobilenet_v2 import *

# Eavise Creations
from ._network_mobile_darknet19 import *
from ._network_mobilenet_yolo import *
from ._network_mobile_yolo_v2 import *
from ._network_mobile_yolo_v2_upsample import *
from ._network_yolo_fusion import *
from ._network_yolo_v2_upsample import *

# Cornernet
from ._network_cornernet import *
from ._network_cornernet_squeeze import *



#
#   Normalized L2 pruning
#   Copyright EAVISE
#

from functools import reduce
import itertools
import logging
import math
import torch
from ._base import *

__all__ = ['L2Pruner']
log = logging.getLogger(__name__)


class L2Pruner(Pruner):
    """ Normalized L2 based pruning. |br|
    This pruner computes the normalized L2 norm per channel
    and prunes the channels with the lowest values.

    The importance of a channel :math:`c_i` of convolution :math:`C` is computed as:

    .. math::
       Importance(c_i) = \\frac{||c_i||_2}{\\sqrt{\\sum_{c_j \\in C} ||c_j||_2^2}}

    Args:
        model (torch.nn.Module): model to prune
        optimizer (torch.optim.Optimizer or None): Optimizer that is used when retraining the network (pass None if not retraining)
        input_dimensions (tuple): Input dimensions to the network
        manner ("soft" or "hard", optional): Whether to perform soft-pruning (replacing channel values with zero) or hard-pruning (deleting channels); Default **"hard"**
        get_parameters (function, optional): function that takes a model and returns the parameters for the optimizer; Default **model.parameters()**

    Note:
        The percentage is approximative. |br|
        If the pruner tells to prune all filters in a convolution,
        we keep the most important and reduce the number of pruned filters by one.
    """
    def prune(self, percentage, prune_manner):
        # Get normalized L2
        l2_norms = []
        num_features = []
        for key, dependency in self.dependencies.items():
            w = dependency.module.weight.view(dependency.module.out_channels, -1)
            l2 = w.pow(2).sum(1) / w.shape[1]
            l2 /= torch.sqrt(l2.pow(2).sum())

            l2_norms.append(l2)
            num_features.append((key, l2.shape[0]))

        # Select filters to prune
        indices = torch.argsort(torch.cat(l2_norms))
        num_pruned = math.floor(indices.numel() * percentage)
        if num_pruned <= 0:
            return num_pruned

        # Convert filters total index to index per convolution
        indices = indices[:num_pruned].tolist()
        indices = map(get_idx_per_conv, indices, itertools.repeat(num_features))
        indices = reduce(get_filter_dict, indices, dict())

        # Prune
        num_features = dict(num_features)
        for key, filter_list in indices.items():
            num = num_features[key]
            while len(filter_list) >= num:
                filter_list = filter_list[:-1]
                num_pruned -= 1

            if len(filter_list) > 0:
                prune_manner(self.dependencies[key], filter_list)

        return num_pruned


def get_idx_per_conv(idx, num_features):
    for name, length in num_features:
        if idx >= length:
            idx -= length
        else:
            return (name, idx)


def get_filter_dict(filter_dict, idx):
    name, idx = idx

    if name not in filter_dict:
        filter_dict[name] = [idx]
    else:
        filter_dict[name].append(idx)

    return filter_dict



"""
Lightnet Data Module |br|
This module contains everything related to pre- and post-processing of your data.
It also has functionality to create datasets from images and annotations that are parseable with brambox_.
"""

from ._dataloading import *
from . import transform



#
#   Darknet Darknet53 model
#   Copyright EAVISE
#

import functools
from collections import OrderedDict
import torch
import torch.nn as nn

import lightnet.network as lnn

__all__ = ['Darknet53']


class Darknet53(lnn.module.Darknet):
    """ Darknet53 implementation :cite:`yolo_v3`.

    Args:
        num_classes (Number, optional): Number of classes; Default **1000**
        input_channels (Number, optional): Number of input channels; Default **3**

    Attributes:
        self.inner_stride: Maximal internal subsampling factor of the network (input dimension should be a multiple of this)
    """
    inner_stride = 32

    def __init__(self, num_classes, input_channels=3):
        """ Network initialisation """
        super().__init__()

        # Parameters
        self.num_classes = num_classes

        # Network
        relu = functools.partial(nn.LeakyReLU, 0.1, inplace=True)
        momentum = 0.01
        self.layers = nn.Sequential(
            OrderedDict([
                ('1_convbatch',         lnn.layer.Conv2dBatchReLU(input_channels, 32, 3, 1, 1, relu=relu, momentum=momentum)),
                ('2_convbatch',         lnn.layer.Conv2dBatchReLU(32, 64, 3, 2, 1, relu=relu, momentum=momentum)),
                ('a_residual',          lnn.layer.Residual(OrderedDict([
                    ('3_convbatch',     lnn.layer.Conv2dBatchReLU(64, 32, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('4_convbatch',     lnn.layer.Conv2dBatchReLU(32, 64, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('5_convbatch',         lnn.layer.Conv2dBatchReLU(64, 128, 3, 2, 1, relu=relu, momentum=momentum)),
                ('b_residual',          lnn.layer.Residual(OrderedDict([
                    ('6_convbatch',     lnn.layer.Conv2dBatchReLU(128, 64, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('7_convbatch',     lnn.layer.Conv2dBatchReLU(64, 128, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('c_residual',          lnn.layer.Residual(OrderedDict([
                    ('8_convbatch',     lnn.layer.Conv2dBatchReLU(128, 64, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('9_convbatch',     lnn.layer.Conv2dBatchReLU(64, 128, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('10_convbatch',        lnn.layer.Conv2dBatchReLU(128, 256, 3, 2, 1, relu=relu, momentum=momentum)),
                ('d_residual',          lnn.layer.Residual(OrderedDict([
                    ('11_convbatch',    lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('12_convbatch',    lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('e_residual',          lnn.layer.Residual(OrderedDict([
                    ('13_convbatch',    lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('14_convbatch',    lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('f_residual',          lnn.layer.Residual(OrderedDict([
                    ('15_convbatch',    lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('16_convbatch',    lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('g_residual',          lnn.layer.Residual(OrderedDict([
                    ('17_convbatch',    lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('18_convbatch',    lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('h_residual',          lnn.layer.Residual(OrderedDict([
                    ('19_convbatch',    lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('20_convbatch',    lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('i_residual',          lnn.layer.Residual(OrderedDict([
                    ('21_convbatch',    lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('22_convbatch',    lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('j_residual',          lnn.layer.Residual(OrderedDict([
                    ('23_convbatch',    lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('24_convbatch',    lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('k_residual',          lnn.layer.Residual(OrderedDict([
                    ('25_convbatch',    lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('26_convbatch',    lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('27_convbatch',        lnn.layer.Conv2dBatchReLU(256, 512, 3, 2, 1, relu=relu, momentum=momentum)),
                ('l_residual',          lnn.layer.Residual(OrderedDict([
                    ('28_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('29_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('m_residual',          lnn.layer.Residual(OrderedDict([
                    ('30_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('31_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('n_residual',          lnn.layer.Residual(OrderedDict([
                    ('32_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('33_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('o_residual',          lnn.layer.Residual(OrderedDict([
                    ('34_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('35_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('p_residual',          lnn.layer.Residual(OrderedDict([
                    ('36_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('37_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('q_residual',          lnn.layer.Residual(OrderedDict([
                    ('38_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('39_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('r_residual',          lnn.layer.Residual(OrderedDict([
                    ('40_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('41_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('s_residual',          lnn.layer.Residual(OrderedDict([
                    ('42_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('43_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('44_convbatch',        lnn.layer.Conv2dBatchReLU(512, 1024, 3, 2, 1, relu=relu, momentum=momentum)),
                ('t_residual',          lnn.layer.Residual(OrderedDict([
                    ('45_convbatch',    lnn.layer.Conv2dBatchReLU(1024, 512, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('46_convbatch',    lnn.layer.Conv2dBatchReLU(512, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('u_residual',          lnn.layer.Residual(OrderedDict([
                    ('47_convbatch',    lnn.layer.Conv2dBatchReLU(1024, 512, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('48_convbatch',    lnn.layer.Conv2dBatchReLU(512, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('v_residual',          lnn.layer.Residual(OrderedDict([
                    ('49_convbatch',    lnn.layer.Conv2dBatchReLU(1024, 512, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('50_convbatch',    lnn.layer.Conv2dBatchReLU(512, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('w_residual',          lnn.layer.Residual(OrderedDict([
                    ('51_convbatch',    lnn.layer.Conv2dBatchReLU(1024, 512, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('52_convbatch',    lnn.layer.Conv2dBatchReLU(512, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))),
                ('53_avgpool',          nn.AdaptiveAvgPool2d(1)),
                ('54_conv',             nn.Conv2d(1024, num_classes, 1, 1, 0)),
                ('55_flatten',          lnn.layer.Flatten()),
            ])
        )



#
#   Base pruning method class
#   Copyright EAVISE
#

from abc import ABC, abstractmethod
import logging
import torch
from .dependency import get_dependency_map, NodeType

__all__ = ['Pruner']
log = logging.getLogger(__name__)


def default_get_parameters(model):
    return model.parameters()


class Pruner(ABC):
    """ Abstract pruning class, which should be inheritted by each different method. |br|
    This class takes care of actually performing the soft- or hard-pruning
    and also updates the optimizer after it deleted channels.

    Args:
        model (torch.nn.Module): model to prune
        input_dimensions (tuple): Input dimensions to the network (width, height, channels) where channels is optional and defaults to 3
        optimizer (torch.optim.Optimizer or None, optional): Optimizer that is used when retraining the network (see Note); default **None**
        manner ("soft" or "hard", optional): Whether to perform soft-pruning (replacing channel values with zero) or hard-pruning (deleting channels); Default **"hard"**
        get_parameters (function, optional): function that takes a model and returns the parameters for the optimizer; Default **model.parameters()**

    Note:
        When performing hard-pruning, the number of parameters of the network changes. |br|
        This means that you need to redefine any object that holds a reference to these parameters.

        When (re-)training a network, this means that you need to recreate a new optimizer, or adapt the existing one.
        By passing an ``optimizer`` to this Pruner class,
        this modification is done for you each time you run the pruner.
        You can also modify which parameters are passed to the optimizer, by passing a ``get_parameters`` function.

        The modifications made to the networks whilst hard-pruning also affect weight loading.
        Lightnet modules thus provide the :func:`~lightnet.network.module.Lightnet.load_pruned` function,
        which allows to correctly adapt a network, according to saved pruned weights.

        These modifications are not necessary for soft-pruning,
        as this technique merely changes the values of the weights.
    """
    def __init__(self, model, input_dimensions, optimizer=None, manner="hard", get_parameters=None):
        self.model = model
        self.optimizer = optimizer
        self.manner = manner
        self._soft_pruned = 0
        self._hard_pruned = 0

        if len(input_dimensions) == 2:
            input_dimensions = (1, 3) + input_dimensions[::-1]
        elif len(input_dimensions) == 3:
            input_dimensions = (1,) + input_dimensions[::-1]
        self.dependencies = get_dependency_map(model, input_dimensions)

        if self.optimizer is None:
            log.warn('Pruner did not get an optimizer, make sure to create a new optimizer each time you train after pruning.')

        if get_parameters is not None:
            self.get_parameters = get_parameters
        else:
            self.get_parameters = default_get_parameters

    def __call__(self, percentage):
        """ Perform pruning.

        Args:
            percentage (float): Percentage of the prunable channels to prune (approximative)

        Returns:
            int: Number of pruned channels
        """
        with torch.no_grad():
            # Delete gradients
            for param in self.model.parameters():
                if hasattr(param, 'grad') and param.grad is not None:
                    param.grad.detach_()
                    param.grad = None

            for buf in self.model.buffers():
                if hasattr(buf, 'grad') and buf.grad is not None:
                    buf.grad.detach_()
                    buf.grad = None

            # Prune
            if self.manner == "soft":
                self._hard_pruned = 0
                prune_count = self._soft_pruned = self.prune(percentage, self.soft_prune)
            else:
                self._soft_pruned = 0
                prune_count = self._hard_pruned = self.prune(percentage, self.hard_prune)
                self._update_optimizer()

            return prune_count

    @abstractmethod
    def prune(self, percentage, prune_manner):
        """ Pruning implementation. |br|
        This method should be implemented by the different methods, and will be called automatically.

        Args:
            percentage (float): Percentage of the prunable channels to prune (approximative)
            prune_manner (function): Function that is used to actually prune (see Note)

        Return:
            int: Number of actually pruned channels

        Note:
            The prune_manner function takes 2 arguments:

            - dependency : item from the dependency-map ``self.dependencies``
            - filter_list : indexes from the channels to prune

            It will then prune the given indexes from the given dependency convolution.
        """
        pass

    @property
    def prunable_channels(self):
        """ Returns the total number of prunable channels left in the network.

        Return:
            int: total number of prunable channels

        Note:
            This property returns the total amount of channels in the prunable convolutions.
            Note that we never prune the last channel of a convolution and thus cannot prune all channels.
        """
        prunable = 0
        for dependency in self.dependencies.values():
            prunable += dependency.module.out_channels

        return prunable

    @property
    def pruned_channels(self):
        """ Returns the number of pruned channels from the last pruning operation. """
        return self._soft_pruned + self._hard_pruned

    @property
    def soft_pruned_channels(self):
        """ Returns the number of soft pruned channels from the last pruning operation. """
        return self._soft_pruned

    @property
    def hard_pruned_channels(self):
        """ Returns the number of hard pruned channels from the last pruning operation. """
        return self._hard_pruned

    def soft_prune(self, conv_node, filter_list, chain=False):
        """ Soft pruning implementation, passed to the prune() function as ``prune_manner`` """
        # Set output tensors to zero; filter_list does not get modified
        if conv_node.module.groups != 1:
            raise NotImplementedError('Grouped Convolution')

        mask = torch.zeros(conv_node.module.weight.shape[0], dtype=bool)
        mask.scatter_(0, torch.tensor(filter_list), True)

        conv_node.module.weight[mask] = 0
        if conv_node.module.bias is not None:
            conv_node.module.bias[mask] = 0

        # Chain modifications
        if chain:
            for c in conv_node.children:
                self._soft_chain(c, filter_list, conv_node)

    def hard_prune(self, conv_node, filter_list, chain=True):
        """ Hard pruning implementation, passed to the prune() function as ``prune_manner`` """
        # Remove filter_list from output tensor; filter_list does not get modified
        if conv_node.module.groups != 1:
            raise NotImplementedError('Grouped Convolution')

        mask = torch.ones(conv_node.module.weight.shape[0], dtype=bool)
        mask.scatter_(0, torch.tensor(filter_list), False)

        conv_node.module.weight = torch.nn.Parameter(conv_node.module.weight[mask])
        if conv_node.module.bias is not None:
            conv_node.module.bias = torch.nn.Parameter(conv_node.module.bias[mask])
        conv_node.module.out_channels -= len(filter_list)

        # Chain modifications
        if chain:
            for c in conv_node.children:
                self._hard_chain(c, filter_list, conv_node)

    def _soft_chain(self, node, filter_list, parent=None):
        if node.type is NodeType.CONV:
            # Set input tensors to zero; filter_list does not get modified
            if node.module.groups != 1:
                # DW-separable
                mask = torch.zeros(node.module.weight.shape[0], dtype=bool)
                mask.scatter_(0, torch.tensor(filter_list), True)
                node.module.weight[mask] = 0
            else:
                mask = torch.zeros(node.module.weight.shape[1], dtype=bool)
                mask.scatter_(0, torch.tensor(filter_list), True)
                node.module.weight[:, mask] = 0
        elif node.type is NodeType.BATCHNORM:
            # Set tensors to zero; filter_list does not get modified
            if node.module.weight is not None:
                mask = torch.zeros(node.module.weight.shape[0], dtype=bool)
                mask.scatter_(0, torch.tensor(filter_list), True)
                node.module.weight[mask] = 0
                if node.module.bias is not None:
                    node.module.bias[mask] = 0

            if node.module.running_mean is not None:
                mask = torch.zeros(node.module.running_mean.shape[0], dtype=bool)
                mask.scatter_(0, torch.tensor(filter_list), True)
                node.module.running_mean[mask] = 0
                if node.module.running_var is not None:
                    node.module.running_var[mask] = 0
        elif node.type is NodeType.CONCAT:
            # Modify filter list according to concat operation
            for p in node.parents:
                if p == parent:
                    break
                else:
                    num = self._get_num_channels(p)
                    filter_list = [f+num for f in filter_list]

        # Recurse
        for c in node.children:
            self._soft_chain(c, filter_list, node)

    def _hard_chain(self, node, filter_list, parent=None):
        if node.type is NodeType.CONV:
            # Remove filter_list from input tensor; filter_list does not get modified
            if node.module.groups != 1:
                # DW-separable
                mask = torch.ones(node.module.weight.shape[0], dtype=bool)
                mask.scatter_(0, torch.tensor(filter_list), False)
                node.module.weight = torch.nn.Parameter(node.module.weight[mask])
                node.module.in_channels -= len(filter_list)
                node.module.out_channels -= len(filter_list)
                node.module.groups -= len(filter_list)
            else:
                mask = torch.ones(node.module.weight.shape[1], dtype=bool)
                mask.scatter_(0, torch.tensor(filter_list), False)
                node.module.weight = torch.nn.Parameter(node.module.weight[:, mask])
                node.module.in_channels -= len(filter_list)
        elif node.type is NodeType.BATCHNORM:
            # Remove filter_list from tensor; filter_list does not get modified
            if node.module.weight is not None:
                mask = torch.ones(node.module.weight.shape[0], dtype=bool)
                mask.scatter_(0, torch.tensor(filter_list), False)
                node.module.weight = torch.nn.Parameter(node.module.weight[mask])
                if node.module.bias is not None:
                    node.module.bias = torch.nn.Parameter(node.module.bias[mask])

            if node.module.running_mean is not None:
                mask = torch.ones(node.module.running_mean.shape[0], dtype=bool)
                mask.scatter_(0, torch.tensor(filter_list), False)
                node.module.running_mean = node.module.running_mean[mask]
                if node.module.running_var is not None:
                    node.module.running_var = node.module.running_var[mask]

            node.module.num_features -= len(filter_list)
        elif node.type is NodeType.CONCAT:
            # Modify filter list according to concat operation
            for p in node.parents:
                if p == parent:
                    break
                else:
                    num = self._get_num_channels(p)
                    filter_list = [f+num for f in filter_list]

        # Recurse
        for c in node.children:
            self._hard_chain(c, filter_list, node)

    def _get_num_channels(self, node):
        if node.type is NodeType.CONV:
            if node.module.groups != 1:
                return node.module.weight.shape[0]
            else:
                return node.module.weight.shape[1]
        if node.type is NodeType.BATCHNORM:
            return node.module.weight.shape[0]
        if node.type is NodeType.CONCAT:
            return sum(self._get_num_channels(p) for p in parents)
        if node.type is NodeType.ELEMW_OP:
            return self._get_num_channels(node.parents[0])

    def _update_optimizer(self):
        if self.optimizer is not None:
            param_groups = list(self.get_parameters(self.model))

            if not isinstance(param_groups[0], dict):
                param_groups = [{'params': param_groups}]

            for old, new in zip(self.optimizer.param_groups, param_groups):
                old.update(new)

    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        get_param = self.get_parameters.__name__ if hasattr(self.get_parameters, '__name__') else self.get_parameters.__class__.__name__
        return f'{self.__class__.__name__} ({self.model.__class__.__name__}, manner={self.manner}, get_parameters={get_param})'



#
#   Test if network weight remapping works
#   Copyright EAVISE
#

import inspect
import pytest
import torch
import lightnet as ln

remaps = [
    # source,                   target,                             remap
    (ln.models.Darknet,         ln.models.TinyYoloV2,               ln.models.TinyYoloV2.remap_darknet),
    (ln.models.Darknet,         ln.models.TinyYoloV3,               ln.models.TinyYoloV3.remap_darknet),
    (ln.models.Darknet19,       ln.models.DYolo,                    ln.models.DYolo.remap_darknet19),
    (ln.models.Darknet19,       ln.models.Yolt,                     ln.models.Yolt.remap_darknet19),
    (ln.models.Darknet19,       ln.models.YoloV2,                   ln.models.YoloV2.remap_darknet19),
    (ln.models.Darknet19,       ln.models.YoloV2Upsample,           ln.models.YoloV2Upsample.remap_darknet19),
    (ln.models.Darknet53,       ln.models.YoloV3,                   ln.models.YoloV3.remap_darknet53),
    (ln.models.MobileDarknet19, ln.models.MobileYoloV2,             ln.models.MobileYoloV2.remap_mobile_darknet19),
    (ln.models.MobileDarknet19, ln.models.MobileYoloV2Upsample,     ln.models.MobileYoloV2.remap_mobile_darknet19),
    (ln.models.MobilenetV1,     ln.models.MobilenetYolo,            ln.models.MobilenetYolo.remap_mobilenet_v1),
]

# Difficult to test (usually remaps from other repos)
remap_skips = [
    ln.models.Cornernet.remap_princeton_vl,
    ln.models.CornernetSqueeze.remap_princeton_vl,
]


@pytest.mark.parametrize('remap', remaps)
def test_remapping(remap, tmp_path):
    # Create networks
    source = remap[0](1000)
    target = remap[1](20)

    # Save weights
    weight_file = str(tmp_path / 'weights.pt')
    source.save(weight_file, remap=remap[2])

    # Check that there are only missing layers and no wrong layers in weight file
    weight_keys = torch.load(weight_file, 'cpu').keys()
    target_keys = target.state_dict().keys()
    assert len(set(weight_keys) - set(target_keys)) == 0

    # Check if loading works
    target.load(weight_file, strict=False)


# All remaps tested?
def test_all_remaps_tested():
    networks = [
        getattr(ln.models, net) for net in dir(ln.models)
        if (inspect.isclass(getattr(ln.models, net)))
        and (issubclass(getattr(ln.models, net), torch.nn.Module))
    ]

    for net in networks:
        net_remaps = [(r, getattr(net, r)) for r in dir(net) if r.startswith('remap')]
        tested_remaps = [r[2] for r in remaps if r[1] == net]

        for remap in net_remaps:
            if remap[1] in remap_skips:
                continue
            if remap[1] not in tested_remaps:
                raise NotImplementedError(f'Remap [{remap[0]}] of Network [{net.__name__}] is not being tested!')



#
#   Cornernet model
#   Copyright EAVISE
#

import functools
import logging
import re
from collections import OrderedDict
import torch
import torch.nn as nn
import lightnet.network as lnn

__all__ = ['Cornernet']
log = logging.getLogger('lightnet.models')


class Cornernet(lnn.module.Lightnet):
    """ Cornernet implementation :cite:`cornernet`.

    .. admonition:: Experimental

       This network implementation is still in development
       and might not be yielding the same results as the official implementation.
       Use at your own risk!

    Args:
        num_classes (Number): Number of classes
        input_channels (Number, optional): Number of input channels; Default **3**
        inference_only (boolean, optional): Whether to load the model purely for inference; Default **False**

    Attributes:
        self.stride: Subsampling factor of the network (input_dim / output_dim)
        self.inner_stride: Maximal internal subsampling factor of the network (input dimension should be a multiple of this)
        self.remap_princeton_vl: Remapping rules for weights from the `official CornerNet implementation <cornernetImpl_>`_.

    .. _cornernetImpl: https://github.com/princeton-vl/CornerNet-Lite
    """
    stride = 4
    inner_stride = 128

    def __init__(self, num_classes, input_channels=3, inference_only=False):
        super().__init__()
        log.experimental(f'"{self.__class__.__name__}" is still in development. Use at your own risk!')

        # Parameters
        self.num_classes = num_classes
        self.input_channels = input_channels
        self.inference_only = inference_only

        # Network
        self.extractor = nn.Sequential(OrderedDict([
            ('1_convbatch',                 lnn.layer.Conv2dBatchReLU(input_channels, 128, 7, 2, 3)),
            ('2_residual',                  self.get_residual(128, 256, stride=2)),
            ('para',                        lnn.layer.ParallelSum(
                lnn.layer.SequentialSelect(['4_convbatch'], False, OrderedDict([
                    ('3_hourglass',         self.get_hourglass(self.get_residual)),
                    ('4_convbatch',         lnn.layer.Conv2dBatchReLU(256, 256, 3, 1, 1)),
                    ('5_conv',              nn.Conv2d(256, 256, 1, 1, 0, bias=False)),
                    ('6_batchnorm',         nn.BatchNorm2d(256)),
                ])),
                nn.Sequential(OrderedDict([
                    ('7_conv',              nn.Conv2d(256, 256, 1, 1, 0, bias=False)),
                    ('8_batchnorm',         nn.BatchNorm2d(256)),
                ])),
                post=nn.ReLU(inplace=True),
            )),
            ('9_residual',                  self.get_residual(256, 256)),
            ('10_hourglass',                self.get_hourglass(self.get_residual)),
            ('11_convbatch',                lnn.layer.Conv2dBatchReLU(256, 256, 3, 1, 1))
        ]))

        self.detector = lnn.layer.ParallelCat(OrderedDict([
            ('topleft',                     nn.Sequential(OrderedDict([
                ('12_corner',               lnn.layer.CornerPool(256, lnn.layer.TopPool, lnn.layer.LeftPool)),
                ('13_convbatch',            lnn.layer.Conv2dBatchReLU(256, 256, 3, 1, 1)),
                ('output',                  lnn.layer.ParallelCat(OrderedDict([
                    ('heatmap',             nn.Sequential(OrderedDict([
                        ('14_conv',         nn.Conv2d(256, 256, 3, 1, 1, bias=True)),
                        ('15_relu',         nn.ReLU(inplace=True)),
                        ('16_conv',         nn.Conv2d(256, num_classes, 1, 1, 0)),
                    ]))),
                    ('embedding',           nn.Sequential(OrderedDict([
                        ('17_conv',         nn.Conv2d(256, 256, 3, 1, 1, bias=True)),
                        ('18_relu',         nn.ReLU(inplace=True)),
                        ('19_conv',         nn.Conv2d(256, 1, 1, 1, 0)),
                    ]))),
                    ('offset',              nn.Sequential(OrderedDict([
                        ('20_conv',         nn.Conv2d(256, 256, 3, 1, 1, bias=True)),
                        ('21_relu',         nn.ReLU(inplace=True)),
                        ('22_conv',         nn.Conv2d(256, 2, 1, 1, 0)),
                    ]))),
                ]))),
            ]))),
            ('bottomright',                 nn.Sequential(OrderedDict([
                ('23_corner',               lnn.layer.CornerPool(256, lnn.layer.BottomPool, lnn.layer.RightPool)),
                ('24_convbatch',            lnn.layer.Conv2dBatchReLU(256, 256, 3, 1, 1)),
                ('output',                  lnn.layer.ParallelCat(OrderedDict([
                    ('heatmap',             nn.Sequential(OrderedDict([
                        ('25_conv',         nn.Conv2d(256, 256, 3, 1, 1, bias=True)),
                        ('26_relu',         nn.ReLU(inplace=True)),
                        ('27_conv',         nn.Conv2d(256, num_classes, 1, 1, 0)),
                    ]))),
                    ('embedding',           nn.Sequential(OrderedDict([
                        ('28_conv',         nn.Conv2d(256, 256, 3, 1, 1, bias=True)),
                        ('29_relu',         nn.ReLU(inplace=True)),
                        ('30_conv',         nn.Conv2d(256, 1, 1, 1, 0)),
                    ]))),
                    ('offset',              nn.Sequential(OrderedDict([
                        ('31_conv',         nn.Conv2d(256, 256, 3, 1, 1, bias=True)),
                        ('32_relu',         nn.ReLU(inplace=True)),
                        ('33_conv',         nn.Conv2d(256, 2, 1, 1, 0)),
                    ]))),
                ]))),
            ]))),
        ]))

        if not self.inference_only:
            self.intermediate = lnn.layer.ParallelCat(OrderedDict([
                ('topleft',                     nn.Sequential(OrderedDict([
                    ('34_corner',               lnn.layer.CornerPool(256, lnn.layer.TopPool, lnn.layer.LeftPool)),
                    ('35_convbatch',            lnn.layer.Conv2dBatchReLU(256, 256, 3, 1, 1)),
                    ('output',                  lnn.layer.ParallelCat(OrderedDict([
                        ('heatmap',             nn.Sequential(OrderedDict([
                            ('36_conv',         nn.Conv2d(256, 256, 3, 1, 1, bias=True)),
                            ('37_relu',         nn.ReLU(inplace=True)),
                            ('38_conv',         nn.Conv2d(256, num_classes, 1, 1, 0)),
                        ]))),
                        ('embedding',           nn.Sequential(OrderedDict([
                            ('39_conv',         nn.Conv2d(256, 256, 3, 1, 1, bias=True)),
                            ('40_relu',         nn.ReLU(inplace=True)),
                            ('41_conv',         nn.Conv2d(256, 1, 1, 1, 0)),
                        ]))),
                        ('offset',              nn.Sequential(OrderedDict([
                            ('42_conv',         nn.Conv2d(256, 256, 3, 1, 1, bias=True)),
                            ('43_relu',         nn.ReLU(inplace=True)),
                            ('44_conv',         nn.Conv2d(256, 2, 1, 1, 0)),
                        ]))),
                    ]))),
                ]))),
                ('bottomright',                 nn.Sequential(OrderedDict([
                    ('45_corner',               lnn.layer.CornerPool(256, lnn.layer.BottomPool, lnn.layer.RightPool)),
                    ('46_convbatch',            lnn.layer.Conv2dBatchReLU(256, 256, 3, 1, 1)),
                    ('output',                  lnn.layer.ParallelCat(OrderedDict([
                        ('heatmap',             nn.Sequential(OrderedDict([
                            ('47_conv',         nn.Conv2d(256, 256, 3, 1, 1, bias=True)),
                            ('48_relu',         nn.ReLU(inplace=True)),
                            ('49_conv',         nn.Conv2d(256, num_classes, 1, 1, 0)),
                        ]))),
                        ('embedding',           nn.Sequential(OrderedDict([
                            ('50_conv',         nn.Conv2d(256, 256, 3, 1, 1, bias=True)),
                            ('51_relu',         nn.ReLU(inplace=True)),
                            ('52_conv',         nn.Conv2d(256, 1, 1, 1, 0)),
                        ]))),
                        ('offset',              nn.Sequential(OrderedDict([
                            ('53_conv',         nn.Conv2d(256, 256, 3, 1, 1, bias=True)),
                            ('54_relu',         nn.ReLU(inplace=True)),
                            ('55_conv',         nn.Conv2d(256, 2, 1, 1, 0)),
                        ]))),
                    ]))),
                ]))),
            ]))

        # Set mode
        if self.inference_only:
            self.eval()

    def forward(self, x):
        return self.forward_train(x) if self.training else self.forward_test(x)

    def forward_train(self, x):
        x = self.extractor(x)
        out1 = self.detector(x)
        out2 = self.intermediate(self.extractor[2][0].selected)
        return out1, out2

    def forward_test(self, x):
        x = self.extractor(x)
        return self.detector(x)

    def train(self, mode=True):
        if mode and self.inference_only:
            raise ValueError("Cannot set training mode for inference_only model")
        return super().train(mode)

    @staticmethod
    def get_residual(in_channels, out_channels, kernel=3, stride=1, padding=1):
        return lnn.layer.Residual(
            nn.Conv2d(in_channels, out_channels, kernel, stride, padding, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_channels, out_channels, kernel, 1, padding, bias=False),
            nn.BatchNorm2d(out_channels),

            skip=None if (in_channels == out_channels) and (stride == 1) else nn.Sequential(
                nn.Conv2d(in_channels, out_channels, 1, stride, 0, bias=False),
                nn.BatchNorm2d(out_channels),
            ),
            post=nn.ReLU(inplace=True),
        )

    @staticmethod
    def get_hourglass(residual):
        return lnn.layer.HourGlass(
            5, [256, 256, 384, 384, 384, 512],
            make_upper=lambda ci, co: nn.Sequential(*[residual(ci, co) for _ in range(2)]),
            make_down1=lambda ci, co: nn.Sequential(residual(ci, co, stride=2), residual(co, co)),
            make_inner=lambda ci, co: nn.Sequential(*[residual(ci, co) for _ in range(4)]),
            make_down2=lambda ci, co: nn.Sequential(residual(ci, ci), residual(ci, co), nn.Upsample(scale_factor=2, mode='nearest'))
        )

    @staticmethod
    def remap_princeton_vl(k):
        remap_extractor = [
            (r'^module.hg.pre.0.conv.(.*)',     r'extractor.1_convbatch.layers.0.\1'),
            (r'^module.hg.pre.0.bn.(.*)',       r'extractor.1_convbatch.layers.1.\1'),
            (r'^module.hg.pre.1.conv1.(.*)',    r'extractor.2_residual.0.\1'),
            (r'^module.hg.pre.1.bn1.(.*)',      r'extractor.2_residual.1.\1'),
            (r'^module.hg.pre.1.conv2.(.*)',    r'extractor.2_residual.3.\1'),
            (r'^module.hg.pre.1.bn2.(.*)',      r'extractor.2_residual.4.\1'),
            (r'^module.hg.pre.1.skip.(.*)',     r'extractor.2_residual.skip.\1'),
            (r'^module.hg.cnvs.0.conv.(.*)',    r'extractor.para.0.4_convbatch.layers.0.\1'),
            (r'^module.hg.cnvs.0.bn.(.*)',      r'extractor.para.0.4_convbatch.layers.1.\1'),
            (r'^module.hg.cnvs.1.conv.(.*)',    r'extractor.11_convbatch.layers.0.\1'),
            (r'^module.hg.cnvs.1.bn.(.*)',      r'extractor.11_convbatch.layers.1.\1'),
            (r'^module.hg.cnvs_.0.0.(.*)',      r'extractor.para.0.5_conv.\1'),
            (r'^module.hg.cnvs_.0.1.(.*)',      r'extractor.para.0.6_batchnorm.\1'),
            (r'^module.hg.inters.0.conv1.(.*)', r'extractor.9_residual.0.\1'),
            (r'^module.hg.inters.0.bn1.(.*)',   r'extractor.9_residual.1.\1'),
            (r'^module.hg.inters.0.conv2.(.*)', r'extractor.9_residual.3.\1'),
            (r'^module.hg.inters.0.bn2.(.*)',   r'extractor.9_residual.4.\1'),
            (r'^module.hg.inters.0.skip.(.*)',  r'extractor.9_residual.skip.\1'),
            (r'^module.hg.inters_.0.0.(.*)',    r'extractor.para.1.7_conv.\1'),
            (r'^module.hg.inters_.0.1.(.*)',    r'extractor.para.1.8_batchnorm.\1'),
        ]
        for r in remap_extractor:
            if re.match(r[0], k) is not None:
                return re.sub(r[0], r[1], k)

        # HOURGLASSES
        if k.startswith('module.hg.hgs'):
            if k[14] == '0':
                nk = 'extractor.para.0.3_hourglass.layers.'
            else:
                nk = 'extractor.10_hourglass.layers.'

            k = (
                k[16:]
                .replace('up1', 'upper')
                .replace('low1', 'down.down1')
                .replace('low3', 'down.down2')
                .replace('conv1', '0')
                .replace('bn1', '1')
                .replace('conv2', '3')
                .replace('bn2', '4')
            )
            k = re.sub(r'low2.(\d+)', r'down.inner.\1', k)
            k = k.replace('low2', 'down.inner.layers')

            return nk+k

        # DETECTION
        _, mod, num, rk = k.split('.', 3)
        corner, mod = mod.split('_')

        if num == '0':
            nk = 'intermediate.'
            num = 22
        else:
            nk = 'detector.'
            num = 0
        if corner == 'tl':
            nk += 'topleft.'
        else:
            nk += 'bottomright.'
            num += 11

        if mod == 'modules':
            remap_detection = [
                (r'^p1_conv1.conv.(.*)',    f'{12+num}_corner.layers.pool.0.0.layers.0.\\1'),
                (r'^p1_conv1.bn.(.*)',      f'{12+num}_corner.layers.pool.0.0.layers.1.\\1'),
                (r'^p2_conv1.conv.(.*)',    f'{12+num}_corner.layers.pool.1.0.layers.0.\\1'),
                (r'^p2_conv1.bn.(.*)',      f'{12+num}_corner.layers.pool.1.0.layers.1.\\1'),
                (r'^p_conv1.(.*)',          f'{12+num}_corner.layers.pool.post.0.\\1'),
                (r'^p_bn1.(.*)',            f'{12+num}_corner.layers.pool.post.1.\\1'),
                (r'^conv1.(.*)',            f'{12+num}_corner.layers.conv.0.\\1'),
                (r'^bn1.(.*)',              f'{12+num}_corner.layers.conv.1.\\1'),
                (r'^conv2.conv.(.*)',       f'{13+num}_convbatch.layers.0.\\1'),
                (r'^conv2.bn.(.*)',         f'{13+num}_convbatch.layers.1.\\1'),
            ]
        elif mod == 'heats':
            remap_detection = [
                (r'0.conv.(.*)',            f'output.heatmap.{14+num}_conv.\\1'),
                (r'1.(.*)',                 f'output.heatmap.{16+num}_conv.\\1'),
            ]
        elif mod == 'tags':
            remap_detection = [
                (r'0.conv.(.*)',            f'output.embedding.{17+num}_conv.\\1'),
                (r'1.(.*)',                 f'output.embedding.{19+num}_conv.\\1'),
            ]
        elif mod == 'offs':
            remap_detection = [
                (r'0.conv.(.*)',            f'output.offset.{20+num}_conv.\\1'),
                (r'1.(.*)',                 f'output.offset.{22+num}_conv.\\1'),
            ]

        for r in remap_detection:
            if re.match(r[0], rk) is not None:
                return nk + re.sub(r[0], r[1], rk)

        log.warn(f'Could not find matching layer for [{k}]')
        return None
